#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Generic Reader for GRIB2 files.

Currently this reader depends on the `pygrib` python package. The `eccodes`
package from ECMWF is preferred, but does not support python 3 at the time
of writing.

"""

import datetime as dt
import logging

import dask.array as da
import numpy as np
import pygrib
import xarray as xr
from pyproj import Proj
from pyresample import geometry

from satpy.dataset import DataQuery
from satpy.readers.core.file_handlers import BaseFileHandler
from satpy.utils import get_legacy_chunk_size

LOG = logging.getLogger(__name__)

CHUNK_SIZE = get_legacy_chunk_size()

CF_UNITS = {
    "none": "1",
}


class GRIBFileHandler(BaseFileHandler):
    """Generic GRIB file handler."""

    def __init__(self, filename, filename_info, filetype_info):
        """Open grib file and do initial message parsing."""
        super(GRIBFileHandler, self).__init__(filename, filename_info, filetype_info)

        self._msg_datasets = {}
        self._start_time = None
        self._end_time = None
        try:
            with pygrib.open(self.filename) as grib_file:
                first_msg = grib_file.message(1)
                last_msg = grib_file.message(grib_file.messages)
                start_time = self._convert_datetime(
                    first_msg, "validityDate", "validityTime")
                end_time = self._convert_datetime(
                    last_msg, "validityDate", "validityTime")
                self._start_time = start_time
                self._end_time = end_time
                if "keys" not in filetype_info:
                    self._analyze_messages(grib_file)
                    self._idx = None
                else:
                    self._create_dataset_ids(filetype_info["keys"])
                    self._idx = pygrib.index(self.filename,
                                             *filetype_info["keys"].keys())
        except (RuntimeError, KeyError):
            raise IOError("Unknown GRIB file format: {}".format(self.filename))

    def _analyze_messages(self, grib_file):
        grib_file.seek(0)
        for idx, msg in enumerate(grib_file):
            msg_id = DataQuery(name=msg["shortName"],
                               level=msg["level"],
                               modifiers=tuple())
            ds_info = {
                "message": idx + 1,
                "name": msg["shortName"],
                "level": msg["level"],
                "file_type": self.filetype_info["file_type"],
            }
            self._msg_datasets[msg_id] = ds_info

    def _create_dataset_ids(self, keys):
        from itertools import product
        ordered_keys = [k for k in keys.keys() if "id_key" in keys[k]]
        for id_vals in product(*[keys[k]["values"] for k in ordered_keys]):
            id_keys = [keys[k]["id_key"] for k in ordered_keys]
            msg_info = dict(zip(ordered_keys, id_vals))
            ds_info = dict(zip(id_keys, id_vals))
            msg_id = DataQuery(**ds_info)
            ds_info = msg_id.to_dict()
            ds_info.update(msg_info)
            ds_info["file_type"] = self.filetype_info["file_type"]
            self._msg_datasets[msg_id] = ds_info

    @staticmethod
    def _convert_datetime(msg, date_key, time_key, date_format="%Y%m%d%H%M"):
        date_str = "{:d}{:04d}".format(msg[date_key], msg[time_key])
        return dt.datetime.strptime(date_str, date_format)

    @property
    def start_time(self):
        """Get start time of this entire file.

        Assumes the first message is the earliest message.

        """
        return self._start_time

    @property
    def end_time(self):
        """Get end time of this entire file.

        Assumes the last message is the latest message.

        """
        return self._end_time

    def available_datasets(self, configured_datasets=None):
        """Automatically determine datasets provided by this file."""
        # previously configured or provided datasets
        # we can't provide any additional information
        for is_avail, ds_info in (configured_datasets or []):
            yield is_avail, ds_info
        # new datasets
        for ds_info in self._msg_datasets.values():
            yield True, ds_info

    def _get_message(self, ds_info):
        with pygrib.open(self.filename) as grib_file:
            if "message" in ds_info:
                msg_num = ds_info["message"]
                msg = grib_file.message(msg_num)
            else:
                msg_keys = self.filetype_info["keys"].keys()
                msg = self._idx(**{k: ds_info[k] for k in msg_keys})[0]
            return msg

    @staticmethod
    def _correct_cyl_minmax_xy(proj_params, min_lon, min_lat, max_lon, max_lat):
        proj = Proj(**proj_params)
        min_x, min_y = proj(min_lon, min_lat)
        max_x, max_y = proj(max_lon, max_lat)
        if max_x <= min_x:
            # wrap around
            # make 180 longitude the prime meridian
            # assuming we are going from 0 to 360 longitude
            proj_params["pm"] = 180
            proj = Proj(**proj_params)
            # recompute x/y extents with this new projection
            min_x, min_y = proj(min_lon, min_lat)
            max_x, max_y = proj(max_lon, max_lat)
        return proj_params, (min_x, min_y, max_x, max_y)

    @staticmethod
    def _get_cyl_minmax_lonlat(lons, lats):
        min_lon = lons[0]
        max_lon = lons[-1]
        min_lat = lats[0]
        max_lat = lats[-1]
        if min_lat > max_lat:
            # lats aren't in the order we thought they were, flip them
            min_lat, max_lat = max_lat, min_lat
        return min_lon, min_lat, max_lon, max_lat

    def _get_cyl_area_info(self, msg, proj_params):
        proj_params["proj"] = "eqc"
        lons = msg["distinctLongitudes"]
        lats = msg["distinctLatitudes"]
        shape = (lats.shape[0], lons.shape[0])
        minmax_lonlat = self._get_cyl_minmax_lonlat(lons, lats)
        proj_params, minmax_xy = self._correct_cyl_minmax_xy(proj_params, *minmax_lonlat)
        extents = self._get_extents(*minmax_xy, shape)
        return proj_params, shape, extents

    @staticmethod
    def _get_extents(min_x, min_y, max_x, max_y, shape):
        half_x = abs((max_x - min_x) / (shape[1] - 1)) / 2.
        half_y = abs((max_y - min_y) / (shape[0] - 1)) / 2.
        return min_x - half_x, min_y - half_y, max_x + half_x, max_y + half_y

    @staticmethod
    def _get_corner_xy(proj_params, lons, lats, scans_positively):
        proj = Proj(**proj_params)
        x, y = proj(lons, lats)
        if scans_positively:
            min_x, min_y = x[0], y[0]
            max_x, max_y = x[3], y[3]
        else:
            min_x, min_y = x[2], y[2]
            max_x, max_y = x[1], y[1]
        return min_x, min_y, max_x, max_y

    @staticmethod
    def _get_corner_lonlat(proj_params, lons, lats):
        # take the corner points only
        lons = lons[([0, 0, -1, -1], [0, -1, 0, -1])]
        lats = lats[([0, 0, -1, -1], [0, -1, 0, -1])]
        # if we have longitudes over 180, assume 0-360
        if (lons > 180).any():
            # make 180 longitude the prime meridian
            proj_params["pm"] = 180
        return proj_params, lons, lats

    def _get_area_info(self, msg, proj_params):
        lats, lons = msg.latlons()
        shape = lats.shape
        scans_positively = (msg.valid_key("jScansPositively") and
                            msg["jScansPositively"] == 1)
        proj_params, lons, lats = self._get_corner_lonlat(
            proj_params, lons, lats)
        minmax_xy = self._get_corner_xy(proj_params, lons, lats, scans_positively)
        extents = self._get_extents(*minmax_xy, shape)
        return proj_params, shape, extents

    @staticmethod
    def _correct_proj_params_over_prime_meridian(proj_params):
        # correct for longitudes over 180
        for lon_param in ["lon_0", "lon_1", "lon_2"]:
            if proj_params.get(lon_param, 0) > 180:
                proj_params[lon_param] -= 360
        return proj_params

    def _area_def_from_msg(self, msg):
        proj_params = msg.projparams.copy()
        proj_params = self._correct_proj_params_over_prime_meridian(proj_params)

        if proj_params["proj"] in ("cyl", "eqc"):
            # eqc projection that goes from 0 to 360
            proj_params, shape, extents = self._get_cyl_area_info(msg, proj_params)
        else:
            proj_params, shape, extents = self._get_area_info(msg, proj_params)

        return geometry.AreaDefinition(
            "on-the-fly grib area",
            "on-the-fly grib area",
            "on-the-fly grib area",
            proj_params,
            shape[1],
            shape[0],
            extents,
        )

    def get_area_def(self, dsid):
        """Get area definition for message.

        If latlong grid then convert to valid eqc grid.

        """
        msg = self._get_message(self._msg_datasets[dsid])
        try:
            return self._area_def_from_msg(msg)
        except (RuntimeError, KeyError):
            raise RuntimeError("Unknown GRIB projection information")

    def get_metadata(self, msg, ds_info):
        """Get metadata."""
        model_time = self._convert_datetime(msg, "dataDate",
                                            "dataTime")
        start_time = self._convert_datetime(msg, "validityDate",
                                            "validityTime")
        end_time = start_time
        try:
            center_description = msg["centreDescription"]
        except (RuntimeError, KeyError):
            center_description = None

        key_dicts = {
            "shortName": "shortName",
            "long_name": "name",
            "pressureUnits": "pressureUnits",
            "typeOfLevel": "typeOfLevel",
            "standard_name": "cfName",
            "units": "units",
            "modelName": "modelName",
            "valid_min": "minimum",
            "valid_max": "maximum",
            "sensor": "modelName"}

        ds_info.update({
            "filename": self.filename,
            "model_time": model_time,
            "centreDescription": center_description,
            "start_time": start_time,
            "end_time": end_time,
            "platform_name": "unknown"})

        for key in key_dicts:
            if key_dicts[key] in msg.keys():
                ds_info[key] = msg[key_dicts[key]]
            else:
                ds_info[key] = "unknown"

        return ds_info

    def get_dataset(self, dataset_id, ds_info):
        """Read a GRIB message into an xarray DataArray."""
        msg = self._get_message(ds_info)
        ds_info = self.get_metadata(msg, ds_info)
        fill = msg["missingValue"]
        data = msg.values.astype(np.float32)
        if msg.valid_key("jScansPositively") and msg["jScansPositively"] == 1:
            data = data[::-1]

        if isinstance(data, np.ma.MaskedArray):
            data = data.filled(np.nan)
            data = da.from_array(data, chunks=CHUNK_SIZE)
        else:
            data[data == fill] = np.nan
            data = da.from_array(data, chunks=CHUNK_SIZE)

        return xr.DataArray(data, attrs=ds_info, dims=("y", "x"))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2016-2023 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""MultiScene object to work with multiple timesteps of satellite data."""
from __future__ import annotations

import copy
import logging
import warnings
from queue import Queue
from threading import Thread
from typing import Callable, Collection, Mapping

import dask.array as da
import numpy as np
import xarray as xr

from satpy.dataset import DataID
from satpy.scene import Scene

try:
    import imageio
except ImportError:
    imageio = None

try:
    from dask.distributed import get_client
except ImportError:
    get_client = None

log = logging.getLogger(__name__)


def _group_datasets_in_scenes(scenes, groups):
    """Group different datasets in multiple scenes by adding aliases.

    Args:
        scenes (Iterable): Scenes to be processed.
        groups (dict): Groups of datasets that shall be treated equally by
            MultiScene. Keys specify the groups, values specify the dataset
            names to be grouped. For example::

                from satpy import DataQuery
                groups = {DataQuery(name='odd'): ['ds1', 'ds3'],
                          DataQuery(name='even'): ['ds2', 'ds4']}

    """
    for scene in scenes:
        grp = _GroupAliasGenerator(scene, groups)
        yield grp.duplicate_datasets_with_group_alias()


class _GroupAliasGenerator:
    """Add group aliases to a scene."""

    def __init__(self, scene, groups):
        """Initialize the alias generator."""
        self.scene = scene.copy()
        self.groups = groups

    def duplicate_datasets_with_group_alias(self):
        """Duplicate datasets to be grouped with a group alias."""
        for group_id, group_members in self.groups.items():
            self._duplicate_dataset_with_group_alias(group_id, group_members)
        return self.scene

    def _duplicate_dataset_with_group_alias(self, group_id, group_members):
        member_ids = self._get_dataset_id_of_group_members_in_scene(group_members)
        if len(member_ids) == 1:
            self._duplicate_dataset_with_different_id(
                dataset_id=member_ids[0],
                alias_id=group_id,
            )
        elif len(member_ids) > 1:
            raise ValueError("Cannot add multiple datasets from a scene "
                             "to the same group")

    def _get_dataset_id_of_group_members_in_scene(self, group_members):
        return [
            self.scene[member].attrs["_satpy_id"]
            for member in group_members if member in self.scene
        ]

    def _duplicate_dataset_with_different_id(self, dataset_id, alias_id):
        dataset = self.scene[dataset_id].copy()
        self._prepare_dataset_for_duplication(dataset, alias_id)
        self.scene[alias_id] = dataset

    def _prepare_dataset_for_duplication(self, dataset, alias_id):
        # Drop all identifier attributes from the original dataset. Otherwise
        # they might invalidate the dataset ID of the alias.
        self._drop_id_attrs(dataset)
        dataset.attrs.update(alias_id.to_dict())

    def _drop_id_attrs(self, dataset):
        for drop_key in self._get_id_attrs(dataset):
            dataset.attrs.pop(drop_key)

    def _get_id_attrs(self, dataset):
        return dataset.attrs["_satpy_id"].to_dict().keys()


class _SceneGenerator(object):
    """Fancy way of caching Scenes from a generator."""

    def __init__(self, scene_gen):
        self._scene_gen = scene_gen
        self._scene_cache = []
        self._dataset_idx = {}
        # this class itself is not an iterator, make one
        self._self_iter = self._create_cached_iter()

    @property
    def first(self):
        """First element in the generator."""
        return next(iter(self))

    def _create_cached_iter(self):
        """Iterate over the provided scenes, caching them for later."""
        for scn in self._scene_gen:
            self._scene_cache.append(scn)
            yield scn

    def __iter__(self):
        """Iterate over the provided scenes, caching them for later."""
        idx = 0
        while True:
            if idx >= len(self._scene_cache):
                try:
                    scn = next(self._self_iter)
                except StopIteration:
                    return
            else:
                scn = self._scene_cache[idx]
            yield scn
            idx += 1

    def __getitem__(self, ds_id):
        """Get a specific dataset from the scenes."""
        for scn in self:
            yield scn.get(ds_id)


class MultiScene(object):
    """Container for multiple `Scene` objects."""

    def __init__(self, scenes=None):
        """Initialize MultiScene and validate sub-scenes.

        Args:
            scenes (Iterable):
                `Scene` objects to operate on (optional)

        .. note::

            If the `scenes` passed to this object are a generator then certain
            operations performed will try to preserve that generator state.
            This may limit what properties or methods are available to the
            user. To avoid this behavior compute the passed generator by
            converting the passed scenes to a list first:
            ``MultiScene(list(scenes))``.

        """
        self._scenes = scenes or []
        scenes = iter(self._scenes)
        self._scene_gen = _SceneGenerator(iter(scenes))
        # if we were originally given a generator-like object then we want to
        # coordinate the loading between _SceneGenerator and _scenes
        # otherwise it doesn't really matter and other operations may prefer
        # a list
        if not isinstance(scenes, (list, tuple)):
            self._scenes = iter(self._scene_gen)

    @property
    def first_scene(self):
        """First Scene of this MultiScene object."""
        return self._scene_gen.first

    @classmethod
    def from_files(  # noqa: D417
            cls,
            files_to_sort: Collection[str],
            reader: str | Collection[str] | None = None,
            ensure_all_readers: bool = False,
            scene_kwargs: Mapping | None = None,
            **kwargs
    ):
        """Create multiple Scene objects from multiple files.

        Args:
            files_to_sort: files to read
            reader: reader or readers to use
            ensure_all_readers: If True, limit to scenes where all
                readers have at least one file.  If False (default), include
                all scenes where at least one reader has at least one file.
            scene_kwargs: additional arguments to pass on to
                :func:`satpy.scene.Scene.__init__` for each created scene.

        This uses the :func:`satpy.readers.core.grouping.group_files` function to group
        files. See this function for more details on additional possible
        keyword arguments.  In particular, it is strongly recommended to pass
        `"group_keys"` when using multiple instruments.

        .. versionadded:: 0.12

        """
        from satpy.readers.core.grouping import group_files
        if scene_kwargs is None:
            scene_kwargs = {}
        file_groups = group_files(files_to_sort, reader=reader, **kwargs)
        if ensure_all_readers:
            warnings.warn(
                "Argument ensure_all_readers is deprecated.  Use "
                "missing='skip' instead.",
                DeprecationWarning,
                stacklevel=2
            )
            file_groups = [fg for fg in file_groups if all(fg.values())]
        scenes = (Scene(filenames=fg, **scene_kwargs) for fg in file_groups)
        return cls(scenes)

    def __iter__(self):
        """Iterate over the provided Scenes once."""
        for scn in self._scenes:
            yield scn

    @property
    def scenes(self):
        """Get list of Scene objects contained in this MultiScene.

        .. note::

            If the Scenes contained in this object are stored in a
            generator (not list or tuple) then accessing this property
            will load/iterate through the generator possibly

        """
        if self.is_generator:
            log.debug("Forcing iteration of generator-like object of Scenes")
            self._scenes = list(self._scenes)
        return self._scenes

    @property
    def is_generator(self):
        """Contained Scenes are stored as a generator."""
        return not isinstance(self._scenes, (list, tuple))

    @property
    def loaded_dataset_ids(self):
        """Union of all Dataset IDs loaded by all children."""
        return set(ds_id for scene in self.scenes for ds_id in scene.keys())

    @property
    def shared_dataset_ids(self):
        """Dataset IDs shared by all children."""
        shared_ids = set(self.scenes[0].keys())
        for scene in self.scenes[1:]:
            shared_ids &= set(scene.keys())
        return shared_ids

    def _all_same_area(self, dataset_ids):
        """Return True if all areas for the provided IDs are equal."""
        all_areas = []
        for ds_id in dataset_ids:
            for scn in self.scenes:
                ds = scn.get(ds_id)
                if ds is None:
                    continue
                all_areas.append(ds.attrs.get("area"))
        all_areas = [area for area in all_areas if area is not None]
        return all(all_areas[0] == area for area in all_areas[1:])

    @property
    def all_same_area(self):
        """Determine if all contained Scenes have the same 'area'."""
        return self._all_same_area(self.loaded_dataset_ids)

    @staticmethod
    def _call_scene_func(gen, func_name, create_new_scene, *args, **kwargs):
        """Abstract method for running a Scene method on each Scene."""
        for scn in gen:
            new_scn = getattr(scn, func_name)(*args, **kwargs)
            if create_new_scene:
                yield new_scn
            else:
                yield scn

    def _generate_scene_func(self, gen, func_name, create_new_scene, *args, **kwargs):
        """Abstract method for running a Scene method on each Scene.

        Additionally, modifies current MultiScene or creates a new one if needed.
        """
        new_gen = self._call_scene_func(gen, func_name, create_new_scene, *args, **kwargs)
        new_gen = new_gen if self.is_generator else list(new_gen)
        if create_new_scene:
            return self.__class__(new_gen)
        self._scene_gen = _SceneGenerator(new_gen)
        self._scenes = iter(self._scene_gen)

    def load(self, *args, **kwargs):
        """Load the required datasets from the multiple scenes."""
        self._generate_scene_func(self._scenes, "load", False, *args, **kwargs)

    def crop(self, *args, **kwargs):
        """Crop the multiscene and return a new cropped multiscene."""
        return self._generate_scene_func(self._scenes, "crop", True, *args, **kwargs)

    def resample(self, destination=None, **kwargs):
        """Resample the multiscene."""
        return self._generate_scene_func(self._scenes, "resample", True, destination=destination, **kwargs)

    def blend(
            self,
            blend_function: Callable[..., xr.DataArray] | None = None
    ) -> Scene:
        """Blend the datasets into one scene.

        Reduce the :class:`MultiScene` to a single :class:`~satpy.scene.Scene`.  Datasets
        occurring in each scene will be passed to a blending
        function, which shall take as input a list of datasets
        (:class:`xarray.DataArray` objects) and shall return a single
        dataset (:class:`xarray.DataArray` object).  The blend method
        then assigns those datasets to the blended scene.

        Blending functions provided in this module are :func:`satpy.multiscene._blend_funcs.stack`
        (the default), :func:`satpy.multiscene._blend_funcs.timeseries`, and
        :func:`satpy.multiscene._blend_funcs.temporal_rgb`, but the Python built-in
        function :func:`sum` also works and may be appropriate for
        some types of data.

        .. note::

            Blending is not currently optimized for generator-based
            MultiScene.

        """
        if blend_function is None:
            # delay importing blend funcs until now in case they aren't used
            from .blend_funcs import stack
            blend_function = stack

        new_scn = Scene()
        common_datasets = self.shared_dataset_ids
        for ds_id in common_datasets:
            datasets = [scn[ds_id] for scn in self.scenes if ds_id in scn]
            new_scn[ds_id] = blend_function(datasets)

        return new_scn

    def group(self, groups):
        """Group datasets from the multiple scenes.

        By default, `MultiScene` only operates on dataset IDs shared by all scenes. Using
        this method you can specify groups of datasets that shall be treated equally
        by `MultiScene`. Even if their dataset IDs differ (for example because the names or
        wavelengths are slightly different).
        Groups can be specified as a dictionary `{group_id: dataset_names}` where the keys
        must be of type `DataQuery`, for example::

            groups={
                DataQuery('my_group', wavelength=(10, 11, 12)): ['IR_108', 'B13', 'C13']
            }
        """
        self._scenes = _group_datasets_in_scenes(self._scenes, groups)

    def _distribute_save_datasets(self, scenes_iter, client, batch_size=1, **kwargs):
        """Distribute save_datasets across a cluster."""
        from satpy.writers.core.compute import split_results

        def load_data(q):
            idx = 0
            while True:
                future_list = q.get()
                if future_list is None:
                    break

                # save_datasets shouldn't be returning anything
                for future in future_list:
                    future.result()
                    log.info("Finished saving %d scenes", idx)
                    idx += 1
                q.task_done()

        input_q = Queue(batch_size if batch_size is not None else 1)
        # set threads to daemon so they are killed if error is raised from main thread
        load_thread = Thread(target=load_data, args=(input_q,), daemon=True)
        load_thread.start()

        for scene in scenes_iter:
            delayeds = scene.save_datasets(compute=False, **kwargs)
            sources, targets, delayeds = split_results([delayeds])
            if len(sources) > 0:
                # TODO Make this work for (source, target) datasets
                # given a target, source combination
                raise NotImplementedError("Distributed save_datasets does not support writers "
                                          "that return (source, target) combinations at this time. Use "
                                          "the non-distributed save_datasets instead.")
            future = client.compute(delayeds)
            input_q.put(future)
        input_q.put(None)

        log.debug("Waiting for child thread to get saved results...")
        load_thread.join()
        log.debug("Child thread died successfully")

    def _simple_save_datasets(self, scenes_iter, **kwargs):
        """Run save_datasets on each Scene."""
        for scn in scenes_iter:
            scn.save_datasets(**kwargs)

    def save_datasets(self, client=True, batch_size=1, **kwargs):
        """Run save_datasets on each Scene.

        Note that some writers may not be multi-process friendly and may
        produce unexpected results or fail by raising an exception. In
        these cases ``client`` should be set to ``False``.
        This is currently a known issue for basic 'geotiff' writer work loads.

        Args:
            batch_size (int): Number of scenes to compute at the same time.
                This only has effect if the `dask.distributed` package is
                installed. This will default to 1. Setting this to 0 or less
                will attempt to process all scenes at once. This option should
                be used with care to avoid memory issues when trying to
                improve performance.
            client (bool or distributed.Client): Dask distributed client
                to use for computation. If this is ``True`` (default) then
                any existing clients will be used.
                If this is ``False`` or ``None`` then a client will not be
                created and ``dask.distributed`` will not be used. If this
                is a dask ``Client`` object then it will be used for
                distributed computation.
            kwargs: Additional keyword arguments to pass to
                    :meth:`~satpy.scene.Scene.save_datasets`.
                    Note ``compute`` can not be provided.

        """
        if "compute" in kwargs:
            raise ValueError("The 'compute' keyword argument can not be provided.")

        client = self._get_client(client=client)

        scenes = iter(self._scenes)
        if client is not None:
            self._distribute_save_datasets(scenes, client, batch_size=batch_size, **kwargs)
        else:
            self._simple_save_datasets(scenes, **kwargs)

    def _get_animation_info(self, all_datasets, filename, fill_value=None):
        """Determine filename and shape of animation to be created."""
        from satpy.enhancements.enhancer import get_enhanced_image

        valid_datasets = [ds for ds in all_datasets if ds is not None]
        first_dataset = valid_datasets[0]
        last_dataset = valid_datasets[-1]
        first_img = get_enhanced_image(first_dataset)
        first_img_data = first_img.finalize(fill_value=fill_value)[0]
        shape = tuple(first_img_data.sizes.get(dim_name)
                      for dim_name in ("y", "x", "bands"))
        if fill_value is None and filename.endswith("gif"):
            log.warning("Forcing fill value to '0' for GIF Luminance images")
            fill_value = 0
            shape = shape[:2]

        attrs = first_dataset.attrs.copy()
        if "end_time" in last_dataset.attrs:
            attrs["end_time"] = last_dataset.attrs["end_time"]
        this_fn = filename.format(**attrs)
        return this_fn, shape, fill_value

    @staticmethod
    def _format_decoration(ds, decorate):
        """Maybe format decoration.

        If the nested dictionary in decorate (argument to ``save_animation``)
        contains a text to be added, format those based on dataset parameters.
        """
        if decorate is None or "decorate" not in decorate:
            return decorate
        deco_local = copy.deepcopy(decorate)
        for deco in deco_local["decorate"]:
            if "text" in deco and "txt" in deco["text"]:
                deco["text"]["txt"] = deco["text"]["txt"].format(**ds.attrs)
        return deco_local

    def _get_single_frame(self, ds, enh_args, fill_value):
        """Get single frame from dataset.

        Yet a single image frame from a dataset.
        """
        from satpy.enhancements.enhancer import get_enhanced_image

        enh_args = enh_args.copy()  # don't change caller's dict!
        if "decorate" in enh_args:
            enh_args["decorate"] = self._format_decoration(
                ds, enh_args["decorate"])
        img = get_enhanced_image(ds, **enh_args)
        data, mode = img.finalize(fill_value=fill_value)
        if data.ndim == 3:
            # assume all other shapes are (y, x)
            # we need arrays grouped by pixel so
            # transpose if needed
            data = data.transpose("y", "x", "bands")
        return data

    def _get_animation_frames(self, all_datasets, shape, fill_value=None,
                              ignore_missing=False, enh_args=None):
        """Create enhanced image frames to save to a file."""
        if enh_args is None:
            enh_args = {}
        for idx, ds in enumerate(all_datasets):
            if ds is None and ignore_missing:
                continue
            elif ds is None:
                log.debug("Missing frame: %d", idx)
                data = da.zeros(shape, dtype=np.uint8, chunks=shape)
                data = xr.DataArray(data)
            else:
                data = self._get_single_frame(ds, enh_args, fill_value)
            yield data.data

    def _get_client(self, client=True):
        """Determine what dask distributed client to use."""
        client = client or None  # convert False/None to None
        if client is True and get_client is None:
            log.debug("'dask.distributed' library was not found, will "
                      "use simple serial processing.")
            client = None
        elif client is True:
            try:
                # get existing client
                client = get_client()
            except ValueError:
                log.warning("No dask distributed client was provided or found, "
                            "but distributed features were requested. Will use simple serial processing.")
                client = None
        return client

    def _distribute_frame_compute(self, writers, frame_keys, frames_to_write, client, batch_size=1):
        """Use ``dask.distributed`` to compute multiple frames at a time."""
        def load_data(frame_gen, q):
            for frame_arrays in frame_gen:
                future_list = client.compute(frame_arrays)
                for frame_key, arr_future in zip(frame_keys, future_list):
                    q.put({frame_key: arr_future})
            q.put(None)

        input_q = Queue(batch_size if batch_size is not None else 1)
        load_thread = Thread(target=load_data, args=(frames_to_write, input_q,))
        load_thread.start()

        while True:
            input_future = input_q.get()
            future_dict = client.gather(input_future)
            if future_dict is None:
                break

            # write the current frame
            # this should only be one element in the dictionary, but this is
            # also the easiest way to get access to the data
            for frame_key, result in future_dict.items():
                # frame_key = rev_future_dict[future]
                w = writers[frame_key]
                w.append_data(result)
            input_q.task_done()

        log.debug("Waiting for child thread...")
        load_thread.join(10)
        if load_thread.is_alive():
            import warnings
            warnings.warn(
                "Background thread still alive after failing to die gracefully",
                stacklevel=3
            )
        else:
            log.debug("Child thread died successfully")

    @staticmethod
    def _simple_frame_compute(writers, frame_keys, frames_to_write):
        """Compute frames the plain dask way."""
        for frame_arrays in frames_to_write:
            for frame_key, product_frame in zip(frame_keys, frame_arrays):
                w = writers[frame_key]
                w.append_data(product_frame.compute())

    def _get_writers_and_frames(
            self, filename, datasets, fill_value, ignore_missing,
            enh_args, imio_args):
        """Get writers and frames.

        Helper function for save_animation.
        """
        scene_gen = self._scene_gen

        first_scene = self.first_scene
        scenes = iter(self._scene_gen)
        info_scenes = [first_scene]
        if "end_time" in filename:
            # if we need the last scene to generate the filename
            # then compute all the scenes so we can figure it out
            log.debug("Generating scenes to compute end_time for filename")
            scenes = list(scenes)
            info_scenes.append(scenes[-1])

        available_ds = [first_scene.get(ds) for ds in first_scene.wishlist]
        available_ds = [DataID.from_dataarray(ds) for ds in available_ds if ds is not None]
        dataset_ids = datasets or available_ds

        if not dataset_ids:
            raise RuntimeError("No datasets found for saving (resampling may be needed to generate composites)")

        writers = {}
        frames = {}
        for dataset_id in dataset_ids:
            if not self.is_generator and not self._all_same_area([dataset_id]):
                raise ValueError("Sub-scene datasets must all be on the same "
                                 "area (see the 'resample' method).")

            all_datasets = scene_gen[dataset_id]
            info_datasets = [scn.get(dataset_id) for scn in info_scenes]
            this_fn, shape, this_fill = self._get_animation_info(info_datasets, filename, fill_value=fill_value)
            data_to_write = self._get_animation_frames(
                all_datasets, shape, this_fill, ignore_missing, enh_args)

            writer = imageio.get_writer(this_fn, **imio_args)
            frames[dataset_id] = data_to_write
            writers[dataset_id] = writer
        return (writers, frames)

    def save_animation(self, filename, datasets=None, fps=10, fill_value=None,
                       batch_size=1, ignore_missing=False, client=True,
                       enh_args=None, **kwargs):
        """Save series of Scenes to movie (MP4) or GIF formats.

        Supported formats are dependent on the `imageio` library and are
        determined by filename extension by default.

        .. note::

            Starting with ``imageio`` 2.5.0, the use of FFMPEG depends on
            a separate ``imageio-ffmpeg`` package.

        By default all datasets available will be saved to individual files
        using the first Scene's datasets metadata to format the filename
        provided. If a dataset is not available from a Scene then a black
        array is used instead (np.zeros(shape)).

        This function can use the ``dask.distributed`` library for improved
        performance by computing multiple frames at a time (see `batch_size`
        option below). If the distributed library is not available then frames
        will be generated one at a time, one product at a time.

        Args:
            filename (str): Filename to save to. Can include python string
                            formatting keys from dataset ``.attrs``
                            (ex. "{name}_{start_time:%Y%m%d_%H%M%S.gif")
            datasets (list): DataIDs to save (default: all datasets)
            fps (int): Frames per second for produced animation
            fill_value (int): Value to use instead creating an alpha band.
            batch_size (int): Number of frames to compute at the same time.
                This only has effect if the `dask.distributed` package is
                installed. This will default to 1. Setting this to 0 or less
                will attempt to process all frames at once. This option should
                be used with care to avoid memory issues when trying to
                improve performance. Note that this is the total number of
                frames for all datasets, so when saving 2 datasets this will
                compute ``(batch_size / 2)`` frames for the first dataset and
                ``(batch_size / 2)`` frames for the second dataset.
            ignore_missing (bool): Don't include a black frame when a dataset
                                   is missing from a child scene.
            client (bool or distributed.Client): Dask distributed client
                to use for computation. If this is ``True`` (default) then
                any existing clients will be used.
                If this is ``False`` or ``None`` then a client will not be
                created and ``dask.distributed`` will not be used. If this
                is a dask ``Client`` object then it will be used for
                distributed computation.
            enh_args (Mapping): Optional, arguments passed to
                :func:`satpy.enhancements.enhancer.get_enhanced_image`.  If this includes a
                keyword "decorate", in any text added
                to the image, string formatting will be applied based on
                dataset attributes.  For example, passing
                ``enh_args={"decorate": {"decorate": [{"text": {"txt":
                "{start_time:%H:%M}"}}]}`` will replace the decorated text
                accordingly.
            kwargs: Additional keyword arguments to pass to
                   `imageio.get_writer`.

        """
        if imageio is None:
            raise ImportError("Missing required 'imageio' library")

        (writers, frames) = self._get_writers_and_frames(
            filename, datasets, fill_value, ignore_missing,
            enh_args, imio_args={"fps": fps, **kwargs})

        client = self._get_client(client=client)
        # get an ordered list of frames
        frame_keys, frames_to_write = list(zip(*frames.items()))
        frames_to_write = zip(*frames_to_write)
        if client is not None:
            self._distribute_frame_compute(writers, frame_keys, frames_to_write, client, batch_size=batch_size)
        else:
            self._simple_frame_compute(writers, frame_keys, frames_to_write)

        for writer in writers.values():
            writer.close()



# Copyright (c) 2022 Satpy developers
#
# satpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# satpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with satpy.  If not, see <http://www.gnu.org/licenses/>.

"""MTG Lightning Imager (LI) Level-2 (L2) unified reader.

This reader supports reading all the products from the LI L2
processing level:

Point products:
  * L2-LE Lightning Events
  * L2-LEF Lightning Events Filtered
  * L2-LFL Lightning Flashes
  * L2-LGR Lightning Groups
Accumulated products:
  * L2-AF Accumulated Flashes
  * L2-AFA Accumulated Flash Area
  * L2-AFR Accumulated Flash Radiance

Per default, the unified LI L2 reader returns the data either as an 1-D array
or as a 2-D array depending on the product type.

Point-based products (LE, LEF, LFL, LGR) are "classic" lightning products
consisting of values with attached latitude and longitude coordinates.
Hence, these products are provided by the reader as 1-D arrays,
with a ``pyresample.geometry.SwathDefinition`` area
attribute containing the points lat-lon coordinates.

Accumulated products (AF, AFA, AFR) are the result of temporal accumulation
of events (e.g. over 30 seconds), and are gridded in the FCI 2km geostationary
projection grid, in order to facilitate the synergistic usage together with FCI.
Compared to the point products, the gridded products also give information
about the spatial extent of the lightning activity.
Hence, these products are provided by the reader as 2-D arrays in the FCI 2km
grid as per intended usage, with a ``pyresample.geometry.AreaDefinition`` area
attribute containing the grid geolocation information.
In this way, the products can directly be overlaid to FCI data.

.. note::

    L2 accumulated products retrieved from the archive
    (that have "ARC" in the filename) contain data for 20 repeat cycles (timesteps) covering
    10 minutes of sensing time. For these files, when loading the main variables
    (``accumulated_flash_area``, ``flash_accumulation``, ``flash_radiance``),
    the reader will cumulate (sum up) the data for the entire sensing period of the file.
    A solution to access easily each timestep is being worked on. See https://github.com/pytroll/satpy/issues/2878
    for possible workarounds in the meanwhile.


If needed, the accumulated products can also be accessed as 1-d array by
setting the reader kwarg ``with_area_definition=False``,
e.g.::

  scn = Scene(filenames=filenames, reader="li_l2_nc", reader_kwargs={'with_area_definition': False})

For both 1-d and 2-d products, the lat-lon coordinates of the points/grid pixels
can be accessed using e.g.
``scn['dataset_name'].attrs['area'].get_lonlats()``.

See the LI L2 Product User Guide `PUG`_ for more information.

.. _PUG: https://www-dr.eumetsat.int/media/49348

"""

import logging

import dask.array as da
import numpy as np
import xarray as xr

from satpy.area import get_area_def
from satpy.readers.core.li_nc import LINCFileHandler
from satpy.utils import get_legacy_chunk_size

logger = logging.getLogger(__name__)
LI_GRID_SHAPE = (5568, 5568)
CHUNK_SIZE = get_legacy_chunk_size()


class LIL2NCFileHandler(LINCFileHandler):
    """Implementation class for the unified LI L2 satpy reader."""

    def __init__(self, filename, filename_info, filetype_info, with_area_definition=True):
        """Initialize LIL2NCFileHandler."""
        super(LIL2NCFileHandler, self).__init__(filename, filename_info, filetype_info)

        if with_area_definition and not self.prod_in_accumulation_grid:
            logger.debug(f"The current product {filetype_info['file_desc']['product_type']} "
                         f"is not an accumulated product so it will not be regridded.")
            self.with_area_def = False
        else:
            self.with_area_def = with_area_definition

    def get_dataset(self, dataset_id, ds_info=None):
        """Get the dataset and apply gridding if requested."""
        data_array = super().get_dataset(dataset_id, ds_info)
        # variable_patterns are compiled to regex patterns
        # hence search variable name from swath_coordinate
        var_with_swath_coord = self.is_var_with_swath_coord(dataset_id)
        if var_with_swath_coord and self.with_area_def:
            data_array = self.get_array_on_fci_grid(data_array)
        else :
            if data_array is not None:
                if not isinstance(data_array.data, da.Array):
                    data_array.data = da.from_array(data_array.data)
        return data_array

    def get_area_def(self, dsid):
        """Compute area definition for a dataset, only supported for accumulated products."""
        var_with_swath_coord = self.is_var_with_swath_coord(dsid)
        if var_with_swath_coord and self.with_area_def:
            return get_area_def("mtg_fci_fdss_2km")

        raise NotImplementedError("Area definition is not supported for non-accumulated products.")

    def is_var_with_swath_coord(self, dsid):
        """Check if the variable corresponding to this dataset is listed as variable with swath coordinates."""
        # since the patterns are compiled to regex we use the search() method below to find matches
        with_swath_coords = any([p.search(dsid["name"]) is not None for p in self.swath_coordinates["patterns"]])
        return with_swath_coords

    def get_array_on_fci_grid(self, data_array: xr.DataArray):
        """Obtain the accumulated products as a (sparse) 2-d array.

        The array has the shape of the FCI 2 km grid (5568x5568px),
        and will have an AreaDefinition attached.
        """
        # Integer values without the application of scale_factor and add_offset
        # hence no projection/index calculation.
        # Note that x and y have origin in the south-west corner of the image
        # and start with index 1.

        rows = self.get_measured_variable("y")
        cols = self.get_measured_variable("x")
        attrs = data_array.attrs

        rows, cols = da.compute(rows, cols)

        # origin is in the south-west corner, so we flip the rows (applying
        # offset of 1 implicitly)
        # And we manually offset the columns by 1 too:
        rows = (LI_GRID_SHAPE[0] - rows.astype(int))
        cols = cols.astype(int) - 1

        # initialise results array with zeros
        data_2d = da.zeros((LI_GRID_SHAPE[0], LI_GRID_SHAPE[1]), dtype=data_array.dtype,
                           chunks=(LI_GRID_SHAPE[0], LI_GRID_SHAPE[1]))

        # insert the data. If a pixel has more than one entry, the values are added up (np.add.at functionality)
        data_2d = da.map_blocks(_np_add_at_wrapper, data_2d, (rows, cols), data_array,
                                dtype=data_array.dtype,
                                chunks=(LI_GRID_SHAPE[0], LI_GRID_SHAPE[1]))
        data_2d = data_2d.astype(np.float32)
        data_2d = da.where(data_2d > 0, data_2d, np.nan)
        xarr = xr.DataArray(da.asarray(data_2d, CHUNK_SIZE), dims=("y", "x"))

        xarr.attrs = attrs
        return xarr


def _np_add_at_wrapper(target_array, indices, data):
    # copy needed for correct computation in-place
    ta = target_array.copy()
    # add.at is not implemented in xarray, so we explicitly need the np.array
    np.add.at(ta, indices, data.values)
    return ta



# -*- coding: utf-8 -*-
# Copyright (c) 2017 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
# type: ignore
"""ScatSat-1 L2B Reader, distributed by Eumetsat in HDF5 format."""

import datetime as dt

import h5py

from satpy.dataset import Dataset
from satpy.readers.core.file_handlers import BaseFileHandler


class SCATSAT1L2BFileHandler(BaseFileHandler):
    """File handler for ScatSat level 2 files, as distributed by Eumetsat in HDF5 format."""

    def __init__(self, filename, filename_info, filetype_info):
        """Initialize the file handler."""
        super(SCATSAT1L2BFileHandler, self).__init__(filename, filename_info, filetype_info)
        self.h5f = h5py.File(self.filename, "r")
        h5data = self.h5f["science_data"]

        self.filename_info["start_time"] = dt.datetime.strptime(
            h5data.attrs["Range Beginning Date"], "%Y-%jT%H:%M:%S.%f")
        self.filename_info["end_time"] = dt.datetime.strptime(
            h5data.attrs["Range Ending Date"], "%Y-%jT%H:%M:%S.%f")

        self.lons = None
        self.lats = None

        self.wind_speed_scale = float(h5data.attrs["Wind Speed Selection Scale"])
        self.wind_direction_scale = float(h5data.attrs["Wind Direction Selection Scale"])
        self.latitude_scale = float(h5data.attrs["Latitude Scale"])
        self.longitude_scale = float(h5data.attrs["Longitude Scale"])

    def get_dataset(self, key, info):
        """Get the dataset."""
        h5data = self.h5f["science_data"]
        stdname = info.get("standard_name")

        if stdname in ["latitude", "longitude"]:

            if self.lons is None or self.lats is None:
                self.lons = h5data["Longitude"][:]*self.longitude_scale
                self.lats = h5data["Latitude"][:]*self.latitude_scale

            if info["standard_name"] == "longitude":
                return Dataset(self.lons, id=key, **info)
            else:
                return Dataset(self.lats, id=key, **info)

        if stdname in ["wind_speed"]:
            windspeed = h5data["Wind_speed_selection"][:, :] * self.wind_speed_scale
            return Dataset(windspeed, id=key, **info)

        if stdname in ["wind_direction"]:
            wind_direction = h5data["Wind_direction_selection"][:, :] * self.wind_direction_scale
            return Dataset(wind_direction, id=key, **info)



# Copyright (c) 2022 Satpy developers
#
# satpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# satpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with satpy.  If not, see <http://www.gnu.org/licenses/>.

"""The atms_l1b_nc reader tests package."""

import datetime as dt

import numpy as np
import pytest
import xarray as xr

from satpy.readers.atms_l1b_nc import AtmsL1bNCFileHandler

# NOTE:
# The following fixtures are not defined in this file, but are used and injected by Pytest:
# - tmp_path


@pytest.fixture
def reader(l1b_file):
    """Return reader of ATMS level1b data."""
    return AtmsL1bNCFileHandler(
        filename=l1b_file,
        filename_info={"creation_time": dt.datetime(2020, 1, 2, 3, 4, 5)},
        filetype_info={"antenna_temperature": "antenna_temp"},
    )


@pytest.fixture
def l1b_file(tmp_path, atms_fake_dataset):
    """Return file path to level1b file."""
    l1b_file_path = tmp_path / "test_file_atms_l1b.nc"
    atms_fake_dataset.to_netcdf(l1b_file_path)
    return l1b_file_path


@pytest.fixture
def atms_fake_dataset():
    """Return fake ATMS dataset."""
    atrack = 2
    xtrack = 3
    channel = 22
    lon = np.full((atrack, xtrack), 1.)
    lat = np.full((atrack, xtrack), 2.)
    sat_azi = np.full((atrack, xtrack), 3.)
    antenna_temp = np.zeros((atrack, xtrack, channel))
    for idx in range(channel):
        antenna_temp[:, :, idx] = 100 + float(idx)
    return xr.Dataset(
        data_vars={
            "antenna_temp": (("atrack", "xtrack", "channel"), antenna_temp),
            "lon": (("atrack", "xtrack"), lon),
            "lat": (("atrack", "xtrack"), lat),
            "sat_azi": (("atrack", "xtrack"), sat_azi),
        },
        attrs={
            "time_coverage_start": "2000-01-02T03:04:05Z",
            "time_coverage_end": "2000-01-02T04:05:06Z",
            "platform": "JPSS-1",
            "instrument": "ATMS",
        },
    )


class TestAtsmsL1bNCFileHandler:
    """Test the AtmsL1bNCFileHandler reader."""

    def test_start_time(self, reader):
        """Test start time."""
        assert reader.start_time == dt.datetime(2000, 1, 2, 3, 4, 5)

    def test_end_time(self, reader):
        """Test end time."""
        assert reader.end_time == dt.datetime(2000, 1, 2, 4, 5, 6)

    def test_sensor(self, reader):
        """Test sensor."""
        assert reader.sensor == "ATMS"

    def test_platform_name(self, reader):
        """Test platform name."""
        assert reader.platform_name == "JPSS-1"

    def test_antenna_temperature(self, reader, atms_fake_dataset):
        """Test antenna temperature."""
        np.testing.assert_array_equal(
            reader.antenna_temperature,
            atms_fake_dataset.antenna_temp.values,
        )

    @pytest.mark.parametrize(("param", "expect"), [
        ("start_time", dt.datetime(2000, 1, 2, 3, 4, 5)),
        ("end_time", dt.datetime(2000, 1, 2, 4, 5, 6)),
        ("platform_name", "JPSS-1"),
        ("sensor", "ATMS"),
    ])
    def test_attrs(self, reader, param, expect):
        """Test attributes."""
        assert reader.attrs[param] == expect

    @pytest.mark.parametrize("dims", [
        ("xtrack", "atrack"),
        ("x", "y"),
    ])
    def test_standardize_dims(self, reader, dims):
        """Test standardize dims."""
        data = xr.DataArray(
            np.arange(6).reshape(2, 3),
            dims=dims,
        )
        standardized = reader._standardize_dims(data)
        assert standardized.dims == ("y", "x")

    def test_drop_coords(self, reader):
        """Test drop coordinates."""
        coords = "dummy"
        data = xr.DataArray(
            np.ones(10),
            dims=("y"),
            coords={coords: 0},
        )
        assert coords in data.coords
        data = reader._drop_coords(data)
        assert coords not in data.coords

    @pytest.mark.parametrize(("param", "expect"), [
        ("start_time", dt.datetime(2000, 1, 2, 3, 4, 5)),
        ("end_time", dt.datetime(2000, 1, 2, 4, 5, 6)),
        ("platform_name", "JPSS-1"),
        ("sensor", "ATMS"),
        ("creation_time", dt.datetime(2020, 1, 2, 3, 4, 5)),
        ("type", "test_data"),
        ("name", "test"),
    ])
    def test_merge_attributes(self, reader, param, expect):
        """Test merge attributes."""
        data = xr.DataArray(
            np.ones(10),
            dims=("y"),
            attrs={"type": "test_data"},
        )
        dataset_info = {"name": "test"}
        data = reader._merge_attributes(data, dataset_info)
        assert data.attrs[param] == expect

    @pytest.mark.parametrize(("param", "expect"), [
        ("1", 100.),
        ("sat_azi", 3.),
    ])
    def test_select_dataset(self, reader, param, expect):
        """Test select dataset."""
        np.testing.assert_array_equal(
            reader._select_dataset(param),
            np.full((2, 3), expect),
        )

    def test_get_dataset(self, reader):
        """Test get dataset."""
        dataset_id = {"name": "1"}
        dataset = reader.get_dataset(dataset_id, {})
        np.testing.assert_array_equal(
            dataset,
            np.full((2, 3), 100.),
        )
        assert dataset.dims == ("y", "x")
        assert dataset.attrs["sensor"] == "ATMS"



# Copyright (c) 2025 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Module for testing the satpy.readers.pace_oci_l1b_nc module."""

import os
import tempfile
from contextlib import suppress
from datetime import datetime

import numpy as np
import pytest
from netCDF4 import Dataset

from satpy import Scene

rng = np.random.default_rng()

n_scan = 10
n_pix = 100
n_blu = 10
n_red = 6
n_swir = 3

blu_wvl = np.array([400.0, 401.0, 402.0, 403.0, 404.0, 405.0, 406.0, 407.0, 408.0, 409.0])
blu_irr = np.array([1102.0, 1103.0, 1104.0, 1105.0, 1106.0, 1107.0, 1108.0, 1109.0, 1110.0, 1111.0])

red_wvl = np.array([600.0, 601.0, 602.0, 603.0, 604.0, 605.0])
red_irr = np.array([102, 104, 107, 99, 102, 59.])

swir_wvl = np.array([1000.0, 1100.0, 1200.0])
swir_bw = np.array([10.0, 20.0, 30.0])
swir_irr = np.array([331.0, 332.0, 333.0])

blu_cdata = rng.uniform(size=(n_blu, n_scan, n_pix), low=0, high=1.04)
red_cdata = rng.uniform(size=(n_red, n_scan, n_pix), low=0, high=1.04)
swir_cdata = rng.uniform(size=(n_swir, n_scan, n_pix), low=0, high=1.04)

sza_data = rng.uniform(size=(n_scan, n_pix), low=0, high=88)
lon_data = rng.uniform(size=(n_scan, n_pix), low=-180, high=180)
lat_data = rng.uniform(size=(n_scan, n_pix), low=-90, high=90)

esd = 0.9989


@pytest.fixture
def temp_nc_file():
    """Create a temporary netCDF file for the tests."""
    temp_dir = tempfile.mkdtemp()
    temp_file = os.path.join(temp_dir, "PACE_OCI.20250320T143701.L1B.V3.nc")

    with Dataset(temp_file, "w", format="NETCDF4") as nc:
        nc.earth_sun_distance_correction = esd

        nc.createDimension("scans", n_scan)
        nc.createDimension("pixels", n_pix)
        nc.createDimension("blue_bands", n_blu)
        nc.createDimension("red_bands", n_red)
        nc.createDimension("swir_bands", n_swir)

        # Global attributes
        nc.time_coverage_start = "2023-01-01T00:29:15.000Z"
        nc.time_coverage_end = "2023-01-01T00:33:11.020Z"

        # Variables
        sensor_band_parameters = nc.createGroup("sensor_band_parameters")
        sensor_band_parameters.createVariable("blue_solar_irradiance", "f4", dimensions=("blue_bands",))
        sensor_band_parameters.createVariable("blue_wavelength", "f4", dimensions=("blue_bands",))

        sensor_band_parameters.createVariable("red_solar_irradiance", "f4", dimensions=("red_bands",))
        sensor_band_parameters.createVariable("red_wavelength", "f4", dimensions=("red_bands",))

        sensor_band_parameters.createVariable("SWIR_solar_irradiance", "f4", dimensions=("swir_bands",))
        sensor_band_parameters.createVariable("SWIR_wavelength", "f4", dimensions=("swir_bands",))
        sensor_band_parameters.createVariable("SWIR_bandpass", "f4", dimensions=("swir_bands",))

        observation_data = nc.createGroup("observation_data")
        observation_data.createVariable("rhot_blue", "f4", ("blue_bands", "scans", "pixels"))
        observation_data.createVariable("rhot_red", "f4", ("red_bands", "scans", "pixels"))
        observation_data.createVariable("rhot_SWIR", "f4", ("swir_bands", "scans", "pixels"))

        geolocation_data = nc.createGroup("geolocation_data")
        geolocation_data.createVariable("solar_zenith", "f4", ("scans", "pixels"))
        geolocation_data.createVariable("latitude", "f4", ("scans", "pixels"))
        geolocation_data.createVariable("longitude", "f4", ("scans", "pixels"))

        # Assign values
        sensor_band_parameters["blue_solar_irradiance"][:] = blu_irr
        sensor_band_parameters["blue_wavelength"][:] = blu_wvl

        sensor_band_parameters["red_solar_irradiance"][:] = red_irr
        sensor_band_parameters["red_wavelength"][:] = red_wvl

        sensor_band_parameters["SWIR_solar_irradiance"][:] = swir_irr
        sensor_band_parameters["SWIR_wavelength"][:] = swir_wvl
        sensor_band_parameters["SWIR_bandpass"][:] = swir_bw

        observation_data["rhot_blue"][:] = blu_cdata
        observation_data["rhot_red"][:] = red_cdata
        observation_data["rhot_SWIR"][:] = swir_cdata

        geolocation_data["solar_zenith"][:] = sza_data
        geolocation_data["longitude"][:] = lon_data
        geolocation_data["latitude"][:] = lat_data

        geolocation_data["latitude"].setncattr("long_name", "Latitudes of pixel locations")
        geolocation_data["latitude"].setncattr("valid_min", "-90.0")
        geolocation_data["latitude"].setncattr("valid_max", "90.0")
        geolocation_data["latitude"].setncattr("units", "degrees_north")
        geolocation_data["latitude"].setncattr("coordinates", "latitude longitude")

        geolocation_data["longitude"].setncattr("long_name", "Longitudes of pixel locations")
        geolocation_data["longitude"].setncattr("valid_min", "-180.0")
        geolocation_data["longitude"].setncattr("valid_max", "180.0")
        geolocation_data["longitude"].setncattr("units", "degrees_east")
        geolocation_data["longitude"].setncattr("coordinates", "latitude longitude")

    yield temp_file

    with suppress(OSError):
        os.remove(temp_file)
        os.rmdir(temp_dir)


@pytest.fixture
def scn(temp_nc_file):
    """Create a Scene object for the tests."""
    return Scene(reader="pace_oci_l1b_nc", filenames=[temp_nc_file])


def test_start_time_property(scn):
    """Test the start_time property."""
    assert scn.start_time == datetime(2023, 1, 1, 0, 29, 15)


def test_end_time_property(scn):
    """Test the end_time property."""
    assert scn.end_time == datetime(2023, 1, 1, 0, 33, 11, 20000)


def test_available_datasets(scn):
    """Test the available_datasets method to check we load correct number dynamically."""
    datasets = list(scn.available_dataset_ids())
    assert len(datasets) == 68


def test_get_dataset_refl(scn):
    """Test getting a dataset with reflectance calibration."""
    cname_b = "chan_blue_400"
    cname_s = "chan_swir_1000"
    scn.load([cname_b, cname_s], calibration="reflectance")
    np.testing.assert_allclose(scn[cname_b], blu_cdata[0, :, :] * 100., rtol=1e-6)
    np.testing.assert_allclose(scn[cname_s], swir_cdata[0, :, :] * 100., rtol=1e-6)

    assert scn[cname_b].attrs["standard_name"] == "toa_bidirectional_reflectance"
    assert scn[cname_s].attrs["standard_name"] == "toa_bidirectional_reflectance"
    assert scn[cname_b].attrs["units"] == "%"
    assert scn[cname_s].attrs["units"] == "%"


def test_get_dataset_radi(scn):
    """Test getting a dataset with radiance calibration."""
    cname_b = "chan_blue_400"
    cname_s = "chan_swir_1000"
    scn.load([cname_b, cname_s], calibration="radiance")

    rad_blu = (blu_cdata[0, :, :] * blu_irr[0] * np.cos(np.radians(sza_data))) / (esd * np.pi)
    rad_swi = (swir_cdata[0, :, :] * swir_irr[0] * np.cos(np.radians(sza_data))) / (esd * np.pi)

    np.testing.assert_allclose(scn[cname_b], rad_blu, rtol=5e-6)
    np.testing.assert_allclose(scn[cname_s], rad_swi, rtol=5e-6)

    assert scn[cname_b].attrs["standard_name"] == "toa_outgoing_radiance_per_unit_wavelength"
    assert scn[cname_s].attrs["standard_name"] == "toa_outgoing_radiance_per_unit_wavelength"
    assert scn[cname_b].attrs["units"] == "W m-2 sr-1 um-1"
    assert scn[cname_s].attrs["units"] == "W m-2 sr-1 um-1"


def test_load_geo(scn):
    """Test loading geolocation data."""
    scn.load(["latitude", "longitude", "solar_zenith_angle"])
    np.testing.assert_allclose(scn["latitude"], lat_data)
    np.testing.assert_allclose(scn["longitude"], lon_data)
    np.testing.assert_allclose(scn["solar_zenith_angle"], sza_data)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2020 Satpy developers
#
# satpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# satpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with satpy.  If not, see <http://www.gnu.org/licenses/>.

"""EUMETSAT EPS-SG Visible/Infrared Imager (VII) Level 2 products reader."""

import logging

import xarray as xr

from satpy.readers.core.vii_nc import ViiNCBaseFileHandler

logger = logging.getLogger(__name__)


class ViiL2NCFileHandler(ViiNCBaseFileHandler):
    """Reader class for VII L2 products in netCDF format."""

    def _perform_orthorectification(self, variable: xr.DataArray, orthorect_data_name: str) -> xr.DataArray:
        """Perform the orthorectification.

        Args:
            variable: DataArray containing the dataset to correct for orthorectification.
            orthorect_data_name: name of the orthorectification correction data in the product.

        Returns:
            array containing the corrected values and all the original metadata.

        """
        try:
            orthorect_data = self[orthorect_data_name]
            variable += orthorect_data
        except KeyError:
            logger.warning("Required dataset %s for orthorectification not available, skipping", orthorect_data_name)
        return variable



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2022-2023 Pytroll developers

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Module for testing the ATMS SDR HDF5 reader."""

import datetime as dt
import os
from unittest import mock

import numpy as np
import pytest

from satpy._config import config_search_paths
from satpy.readers.atms_sdr_hdf5 import ATMS_CHANNEL_NAMES
from satpy.readers.core.loading import load_reader
from satpy.readers.core.viirs_atms_sdr import DATASET_KEYS
from satpy.tests.reader_tests.test_hdf5_utils import FakeHDF5FileHandler

DEFAULT_FILE_DTYPE = np.uint16
DEFAULT_FILE_SHAPE = (1, 96)
# Mimicking one scan line of data
DEFAULT_FILE_DATA = np.arange(DEFAULT_FILE_SHAPE[0] * DEFAULT_FILE_SHAPE[1],
                              dtype=DEFAULT_FILE_DTYPE).reshape(DEFAULT_FILE_SHAPE)
DEFAULT_FILE_FACTORS = np.array([2.0, 1.0], dtype=np.float32)


class FakeHDF5_ATMS_SDR_FileHandler(FakeHDF5FileHandler):
    """Swap-in HDF5 File Handler."""

    _num_test_granules = 1
    _num_scans_per_gran = [12]
    _num_of_bands = 22

    def __init__(self, filename, filename_info, filetype_info, include_factors=True):
        """Create fake file handler."""
        self.include_factors = include_factors
        super().__init__(filename, filename_info, filetype_info)

    @staticmethod
    def _add_basic_metadata_to_file_content(file_content, filename_info, num_grans):
        start_time = filename_info["start_time"]
        end_time = filename_info["end_time"].replace(year=start_time.year,
                                                     month=start_time.month,
                                                     day=start_time.day)
        begin_date = start_time.strftime("%Y%m%d")
        begin_time = start_time.strftime("%H%M%S.%fZ")
        ending_date = end_time.strftime("%Y%m%d")
        ending_time = end_time.strftime("%H%M%S.%fZ")
        new_file_content = {
            "{prefix2}/attr/AggregateNumberGranules": num_grans,
            "{prefix2}/attr/AggregateBeginningDate": begin_date,
            "{prefix2}/attr/AggregateBeginningTime": begin_time,
            "{prefix2}/attr/AggregateEndingDate": ending_date,
            "{prefix2}/attr/AggregateEndingTime": ending_time,
            "{prefix2}/attr/G-Ring_Longitude": np.array([0.0, 0.1, 0.2, 0.3]),
            "{prefix2}/attr/G-Ring_Latitude": np.array([0.0, 0.1, 0.2, 0.3]),
            "{prefix2}/attr/AggregateBeginningOrbitNumber": "{0:d}".format(filename_info["orbit"]),
            "{prefix2}/attr/AggregateEndingOrbitNumber": "{0:d}".format(filename_info["orbit"]),
            "{prefix1}/attr/Instrument_Short_Name": "ATMS",
            "/attr/Platform_Short_Name": "J01",
        }
        file_content.update(new_file_content)

    def _add_granule_specific_info_to_file_content(self, file_content, dataset_group,
                                                   num_granules, num_scans_per_granule, gran_group_prefix):
        lons_lists = self._get_per_granule_lons()
        lats_lists = self._get_per_granule_lats()
        file_content["{prefix3}/NumberOfScans"] = np.array([1] * num_granules)
        for granule_idx in range(num_granules):
            prefix_gran = "{prefix}/{dataset_group}_Gran_{idx}".format(prefix=gran_group_prefix,
                                                                       dataset_group=dataset_group,
                                                                       idx=granule_idx)
            num_scans = num_scans_per_granule[granule_idx]
            file_content[prefix_gran + "/attr/N_Number_Of_Scans"] = num_scans
            file_content[prefix_gran + "/attr/G-Ring_Longitude"] = lons_lists[granule_idx]
            file_content[prefix_gran + "/attr/G-Ring_Latitude"] = lats_lists[granule_idx]

    @staticmethod
    def _get_per_granule_lons():
        return [
            np.array(
                [
                    50.51393, 49.566296, 48.865967, 18.96082, -4.0238385,
                    -7.05221, -10.405702, 14.638646
                ],
                dtype=np.float32),
            np.array(
                [
                    53.52594, 51.685738, 50.439102, 14.629087, -10.247547,
                    -13.951393, -18.256989, 8.36572
                ],
                dtype=np.float32),
            np.array(
                [
                    59.386833, 55.770416, 53.38952, 8.353765, -18.062435,
                    -22.608992, -27.867302, -1.3537619
                ],
                dtype=np.float32),
            np.array(
                [
                    72.50243, 64.17125, 59.15234, -1.3654504, -27.620953,
                    -33.091743, -39.28113, -17.749891
                ],
                dtype=np.float32)
        ]

    @staticmethod
    def _get_per_granule_lats():
        return [
            np.array(
                [
                    67.969505, 65.545685, 63.103046, 61.853905, 55.169273,
                    57.062447, 58.86063, 66.495514
                ],
                dtype=np.float32),
            np.array(
                [
                    72.74879, 70.2493, 67.84738, 66.49691, 58.77254,
                    60.465942, 62.11525, 71.08249
                ],
                dtype=np.float32),
            np.array(
                [
                    77.393425, 74.977875, 72.62976, 71.083435, 62.036346,
                    63.465122, 64.78075, 75.36842
                ],
                dtype=np.float32),
            np.array(
                [
                    81.67615, 79.49934, 77.278656, 75.369415, 64.72178,
                    65.78417, 66.66166, 79.00025
                ],
                dtype=np.float32),
        ]

    def _add_data_info_to_file_content(self, file_content, filename, data_var_prefix, num_grans):
        # ATMS SDR files always produce data with 12 scans per granule even if there are less? FIXME!
        total_rows = DEFAULT_FILE_SHAPE[0] * 12 * num_grans
        new_shape = (total_rows, DEFAULT_FILE_SHAPE[1], self._num_of_bands)
        key = "BrightnessTemperature"
        key = data_var_prefix + "/" + key
        file_content[key] = np.repeat(DEFAULT_FILE_DATA.copy(), 12 * num_grans, axis=0)
        file_content[key] = np.repeat(file_content[key][:, :, np.newaxis], self._num_of_bands, axis=2)
        file_content[key + "/shape"] = new_shape
        if self.include_factors:
            file_content[key + "Factors"] = np.repeat(
                DEFAULT_FILE_FACTORS.copy()[None, :], num_grans, axis=0).ravel()

    @staticmethod
    def _add_geolocation_info_to_file_content(file_content, filename, data_var_prefix, num_grans):
        # ATMS SDR files always produce data with 12 scans per granule even if there are less? FIXME!
        total_rows = DEFAULT_FILE_SHAPE[0] * 12 * num_grans
        new_shape = (total_rows, DEFAULT_FILE_SHAPE[1])

        lon_data = np.linspace(15, 55, DEFAULT_FILE_SHAPE[1]).astype(DEFAULT_FILE_DTYPE)
        lat_data = np.linspace(55, 75, DEFAULT_FILE_SHAPE[1]).astype(DEFAULT_FILE_DTYPE)

        for k in ["Latitude"]:
            k = data_var_prefix + "/" + k
            file_content[k] = lat_data
            file_content[k] = np.repeat([file_content[k]], total_rows, axis=0)
            file_content[k + "/shape"] = new_shape
        for k in ["Longitude"]:
            k = data_var_prefix + "/" + k
            file_content[k] = lon_data
            file_content[k] = np.repeat([file_content[k]], total_rows, axis=0)
            file_content[k + "/shape"] = new_shape

        angles = ["SolarZenithAngle",
                  "SolarAzimuthAngle",
                  "SatelliteZenithAngle",
                  "SatelliteAzimuthAngle"]
        for k in angles:
            k = data_var_prefix + "/" + k
            file_content[k] = lon_data  # close enough to SZA
            file_content[k] = np.repeat([file_content[k]], total_rows, axis=0)
            file_content[k + "/shape"] = new_shape

    @staticmethod
    def _add_geo_ref(file_content, filename):
        geo_prefix = "GATMO"
        file_content["/attr/N_GEO_Ref"] = geo_prefix + filename[5:]

    @staticmethod
    def _convert_numpy_content_to_dataarray(final_content):
        import dask.array as da
        from xarray import DataArray
        for key, val in final_content.items():
            if isinstance(val, np.ndarray):
                val = da.from_array(val, chunks=val.shape)
                if val.ndim > 2:
                    final_content[key] = DataArray(val, dims=("y", "x", "z"))
                elif val.ndim > 1:
                    final_content[key] = DataArray(val, dims=("y", "x"))
                else:
                    final_content[key] = DataArray(val)

    def get_test_content(self, filename, filename_info, filetype_info):
        """Mimic reader input file content."""
        final_content = {}
        for dataset in self.datasets:
            dataset_group = DATASET_KEYS[dataset]
            prefix1 = "Data_Products/{dataset_group}".format(dataset_group=dataset_group)
            prefix2 = "{prefix}/{dataset_group}_Aggr".format(prefix=prefix1, dataset_group=dataset_group)
            prefix3 = "All_Data/{dataset_group}_All".format(dataset_group=dataset_group)

            file_content = {}
            self._add_basic_metadata_to_file_content(file_content, filename_info, self._num_test_granules)
            self._add_granule_specific_info_to_file_content(file_content, dataset_group,
                                                            self._num_test_granules, self._num_scans_per_gran,
                                                            prefix1)
            self._add_geo_ref(file_content, filename)

            for k, v in list(file_content.items()):
                file_content[k.format(prefix1=prefix1, prefix2=prefix2, prefix3=prefix3)] = v

            if filename[:5] in ["SATMS", "TATMS"]:
                self._add_data_info_to_file_content(file_content, filename, prefix3,
                                                    self._num_test_granules)
            elif filename[0] == "G":
                self._add_geolocation_info_to_file_content(file_content, filename, prefix3,
                                                           self._num_test_granules)
            final_content.update(file_content)
        self._convert_numpy_content_to_dataarray(final_content)
        return final_content


class TestATMS_SDR_Reader:
    """Test ATMS SDR Reader."""

    yaml_file = "atms_sdr_hdf5.yaml"

    def _assert_bt_properties(self, data_arr, num_scans=1, with_area=True):
        assert np.issubdtype(data_arr.dtype, np.float32)

        assert data_arr.attrs["calibration"] == "brightness_temperature"
        assert data_arr.attrs["units"] == "K"
        assert data_arr.attrs["rows_per_scan"] == num_scans
        if with_area:
            assert "area" in data_arr.attrs
            assert data_arr.attrs["area"] is not None
            assert data_arr.attrs["area"].shape == data_arr.shape
        else:
            assert "area" not in data_arr.attrs

    def setup_method(self):
        """Wrap HDF5 file handler with our own fake handler."""
        from satpy.readers.core.viirs_atms_sdr import JPSS_SDR_FileHandler

        self.reader_configs = config_search_paths(os.path.join("readers", self.yaml_file))
        # http://stackoverflow.com/questions/12219967/how-to-mock-a-base-class-with-python-mock-library
        self.p = mock.patch.object(JPSS_SDR_FileHandler, "__bases__", (FakeHDF5_ATMS_SDR_FileHandler,))
        self.fake_handler = self.p.start()
        self.p.is_local = True

    def teardown_method(self):
        """Stop wrapping the HDF5 file handler."""
        self.p.stop()

    def test_init(self):
        """Test basic init with no extra parameters."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            "/path/to/atms/sdr/data/SATMS_j01_d20221220_t0910240_e0921356_b26361_c20221220100456348770_cspp_dev.h5",
        ])
        assert len(loadables) == 1
        r.create_filehandlers(loadables)
        # make sure we have some files
        assert r.file_handlers

    def test_init_start_end_time(self):
        """Test basic init with start and end times around the start/end times of the provided file."""
        r = load_reader(self.reader_configs,
                        filter_parameters={
                            "start_time": dt.datetime(2022, 12, 19),
                            "end_time": dt.datetime(2022, 12, 21)
                        })
        loadables = r.select_files_from_pathnames([
            "SATMS_j01_d20221220_t0910240_e0921356_b26361_c20221220100456348770_cspp_dev.h5",
        ])
        assert len(loadables) == 1
        r.create_filehandlers(loadables)
        # make sure we have some files
        assert r.file_handlers

    @pytest.mark.parametrize(("files", "expected"),
                             [(["SATMS_j01_d20221220_t0910240_e0921356_b26361_c20221220100456348770_cspp_dev.h5",
                                "GATMO_j01_d20221220_t0910240_e0921356_b26361_c20221220100456680030_cspp_dev.h5"],
                               True),
                              (["SATMS_j01_d20221220_t0910240_e0921356_b26361_c20221220100456348770_cspp_dev.h5", ],
                               False)]
                             )
    def test_load_all_bands(self, files, expected):
        """Load brightness temperatures for all 22 ATMS channels, with/without geolocation."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames(files)
        r.create_filehandlers(loadables)
        ds = r.load(ATMS_CHANNEL_NAMES)
        assert len(ds) == 22
        for d in ds.values():
            self._assert_bt_properties(d, with_area=expected)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2021 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Demo data download helper functions for AHI HSD data."""
import os

from satpy import config


def download_typhoon_surigae_ahi(base_dir=None,
                                 channels=(1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16),
                                 segments=(1, 2, 3, 4, 5, 6, 7, 8, 9, 10)):
    """Download Himawari 8 data.

    This scene shows the Typhoon Surigae.
    """
    import s3fs
    base_dir = base_dir or config.get("demo_data_dir", ".")
    channel_resolution = {1: 10,
                          2: 10,
                          3: 5,
                          4: 10}
    data_files = []
    for channel in channels:
        resolution = channel_resolution.get(channel, 20)
        for segment in segments:
            data_files.append(f"HS_H08_20210417_0500_B{channel:02d}_FLDK_R{resolution:02d}_S{segment:02d}10.DAT.bz2")

    subdir = os.path.join(base_dir, "ahi_hsd", "20210417_0500_typhoon_surigae")
    os.makedirs(subdir, exist_ok=True)
    fs = s3fs.S3FileSystem(anon=True)

    result = []
    for filename in data_files:
        destination_filename = os.path.join(subdir, filename)
        result.append(destination_filename)
        if os.path.exists(destination_filename):
            continue
        to_get = "noaa-himawari8/AHI-L1b-FLDK/2021/04/17/0500/" + filename
        fs.get_file(to_get, destination_filename)

    return result



# Copyright (c) 2010-2023 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Unit tests for data access methods and properties of the Scene class."""
import math

import numpy as np
import pytest
import xarray as xr
from dask import array as da

from satpy import Scene
from satpy.dataset.dataid import default_id_keys_config
from satpy.tests.utils import FAKE_FILEHANDLER_END, FAKE_FILEHANDLER_START, make_cid, make_dataid

# NOTE:
# The following fixtures are not defined in this file, but are used and injected by Pytest:
# - include_test_etc


@pytest.mark.usefixtures("include_test_etc")
class TestDataAccessMethods:
    """Test the scene class."""

    @pytest.mark.parametrize(
        ("reader", "filenames", "exp_sensors"),
        [
            ("fake1", ["fake1_1.txt"], {"fake_sensor"}),
            (None, {"fake1": ["fake1_1.txt"], "fake2_1ds": ["fake2_1ds_1.txt"]}, {"fake_sensor", "fake_sensor2"}),
        ]
    )
    def test_sensor_names_readers(self, reader, filenames, exp_sensors):
        """Test that Scene sensor_names handles different cases properly."""
        scene = Scene(reader=reader, filenames=filenames)
        assert scene.start_time == FAKE_FILEHANDLER_START
        assert scene.end_time == FAKE_FILEHANDLER_END
        assert scene.sensor_names == exp_sensors

    @pytest.mark.parametrize(
        ("include_reader", "added_sensor", "exp_sensors"),
        [
            (False, "my_sensor", {"my_sensor"}),
            (True, "my_sensor", {"my_sensor", "fake_sensor"}),
            (False, {"my_sensor"}, {"my_sensor"}),
            (True, {"my_sensor"}, {"my_sensor", "fake_sensor"}),
            (False, {"my_sensor1", "my_sensor2"}, {"my_sensor1", "my_sensor2"}),
            (True, {"my_sensor1", "my_sensor2"}, {"my_sensor1", "my_sensor2", "fake_sensor"}),
        ]
    )
    def test_sensor_names_added_datasets(self, include_reader, added_sensor, exp_sensors):
        """Test that Scene sensor_names handles contained sensors properly."""
        if include_reader:
            scene = Scene(reader="fake1", filenames=["fake1_1.txt"])
        else:
            scene = Scene()

        scene["my_ds"] = xr.DataArray([], attrs={"sensor": added_sensor})
        assert scene.sensor_names == exp_sensors

    def test_iter(self):
        """Test iteration over the scene."""
        scene = Scene()
        scene["1"] = xr.DataArray(np.arange(5))
        scene["2"] = xr.DataArray(np.arange(5))
        scene["3"] = xr.DataArray(np.arange(5))
        for x in scene:
            assert isinstance(x, xr.DataArray)

    def test_iter_by_area_swath(self):
        """Test iterating by area on a swath."""
        from pyresample.geometry import SwathDefinition
        scene = Scene()
        sd = SwathDefinition(lons=np.arange(5), lats=np.arange(5))
        scene["1"] = xr.DataArray(np.arange(5), attrs={"area": sd})
        scene["2"] = xr.DataArray(np.arange(5), attrs={"area": sd})
        scene["3"] = xr.DataArray(np.arange(5))
        for area_obj, ds_list in scene.iter_by_area():
            ds_list_names = set(ds["name"] for ds in ds_list)
            if area_obj is sd:
                assert ds_list_names == {"1", "2"}
            else:
                assert area_obj is None
                assert ds_list_names == {"3"}

    def test_bad_setitem(self):
        """Test setting an item wrongly."""
        scene = Scene()
        with pytest.raises(ValueError, match="Key must be a DataID when value is not an xarray DataArray or dict"):
            scene.__setitem__("1", np.arange(5))

    def test_setitem(self):
        """Test setting an item."""
        from satpy.tests.utils import make_dataid
        scene = Scene()
        scene["1"] = ds1 = xr.DataArray(np.arange(5))
        expected_id = make_cid(**ds1.attrs)
        assert set(scene._datasets.keys()) == {expected_id}
        assert set(scene._wishlist) == {expected_id}

        did = make_dataid(name="oranges")
        scene[did] = ds1
        assert "oranges" in scene
        nparray = np.arange(5*5).reshape(5, 5)
        with pytest.raises(ValueError, match="Key must be a DataID when value is not an xarray DataArray or dict"):
            scene["apples"] = nparray
        assert "apples" not in scene
        did = make_dataid(name="apples")
        scene[did] = nparray
        assert "apples" in scene

    def test_getitem(self):
        """Test __getitem__ with names only."""
        scene = Scene()
        scene["1"] = ds1 = xr.DataArray(np.arange(5))
        scene["2"] = ds2 = xr.DataArray(np.arange(5))
        scene["3"] = ds3 = xr.DataArray(np.arange(5))
        assert scene["1"] is ds1
        assert scene["2"] is ds2
        assert scene["3"] is ds3
        pytest.raises(KeyError, scene.__getitem__, "4")
        assert scene.get("3") is ds3
        assert scene.get("4") is None

    def test_getitem_modifiers(self):
        """Test __getitem__ with names and modifiers."""
        # Return least modified item
        scene = Scene()
        scene["1"] = ds1_m0 = xr.DataArray(np.arange(5))
        scene[make_dataid(name="1", modifiers=("mod1",))
              ] = xr.DataArray(np.arange(5))
        assert scene["1"] is ds1_m0
        assert len(list(scene.keys())) == 2

        scene = Scene()
        scene["1"] = ds1_m0 = xr.DataArray(np.arange(5))
        scene[make_dataid(name="1", modifiers=("mod1",))
              ] = xr.DataArray(np.arange(5))
        scene[make_dataid(name="1", modifiers=("mod1", "mod2"))
              ] = xr.DataArray(np.arange(5))
        assert scene["1"] is ds1_m0
        assert len(list(scene.keys())) == 3

        scene = Scene()
        scene[make_dataid(name="1", modifiers=("mod1", "mod2"))
              ] = ds1_m2 = xr.DataArray(np.arange(5))
        scene[make_dataid(name="1", modifiers=("mod1",))
              ] = ds1_m1 = xr.DataArray(np.arange(5))
        assert scene["1"] is ds1_m1
        assert scene[make_dataid(name="1", modifiers=("mod1", "mod2"))] is ds1_m2
        pytest.raises(KeyError, scene.__getitem__,
                      make_dataid(name="1", modifiers=tuple()))
        assert len(list(scene.keys())) == 2

    def test_getitem_slices(self):
        """Test __getitem__ with slices."""
        from pyresample.geometry import AreaDefinition, SwathDefinition
        scene1 = Scene()
        scene2 = Scene()
        area_def = AreaDefinition(
            "test",
            "test",
            "test",
            "+proj=lcc +datum=WGS84 +ellps=WGS84 +lon_0=-95. +lat_0=25 +lat_1=25 +units=m +no_defs",
            200,
            400,
            (-1000., -1500., 1000., 1500.),
        )
        swath_def = SwathDefinition(lons=np.zeros((5, 10)),
                                    lats=np.zeros((5, 10)))
        scene1["1"] = scene2["1"] = xr.DataArray(np.zeros((5, 10)))
        scene1["2"] = scene2["2"] = xr.DataArray(np.zeros((5, 10)),
                                                 dims=("y", "x"))
        scene1["3"] = xr.DataArray(np.zeros((5, 10)), dims=("y", "x"),
                                   attrs={"area": area_def})
        anc_vars = [xr.DataArray(np.ones((5, 10)),
                                 attrs={"name": "anc_var", "area": area_def})]
        attrs = {"ancillary_variables": anc_vars, "area": area_def}
        scene1["3a"] = xr.DataArray(np.zeros((5, 10)),
                                    dims=("y", "x"),
                                    attrs=attrs)
        scene2["4"] = xr.DataArray(np.zeros((5, 10)), dims=("y", "x"),
                                   attrs={"area": swath_def})
        anc_vars = [xr.DataArray(np.ones((5, 10)),
                                 attrs={"name": "anc_var", "area": swath_def})]
        attrs = {"ancillary_variables": anc_vars, "area": swath_def}
        scene2["4a"] = xr.DataArray(np.zeros((5, 10)),
                                    dims=("y", "x"),
                                    attrs=attrs)
        new_scn1 = scene1[2:5, 2:8]
        new_scn2 = scene2[2:5, 2:8]
        for new_scn in [new_scn1, new_scn2]:
            # datasets without an area don't get sliced
            assert new_scn["1"].shape == (5, 10)
            assert new_scn["2"].shape == (5, 10)

        assert new_scn1["3"].shape == (3, 6)
        assert "area" in new_scn1["3"].attrs
        assert new_scn1["3"].attrs["area"].shape == (3, 6)
        assert new_scn1["3a"].shape == (3, 6)
        a_var = new_scn1["3a"].attrs["ancillary_variables"][0]
        assert a_var.shape == (3, 6)

        assert new_scn2["4"].shape == (3, 6)
        assert "area" in new_scn2["4"].attrs
        assert new_scn2["4"].attrs["area"].shape == (3, 6)
        assert new_scn2["4a"].shape == (3, 6)
        a_var = new_scn2["4a"].attrs["ancillary_variables"][0]
        assert a_var.shape == (3, 6)

    def test_contains(self):
        """Test contains."""
        scene = Scene()
        scene["1"] = xr.DataArray(np.arange(5),
                                  attrs={"wavelength": (0.1, 0.2, 0.3),
                                         "_satpy_id_keys": default_id_keys_config})
        assert "1" in scene
        assert 0.15 in scene
        assert "2" not in scene
        assert 0.31 not in scene

        scene = Scene()
        scene["blueberry"] = xr.DataArray(np.arange(5))
        scene["blackberry"] = xr.DataArray(np.arange(5))
        scene["strawberry"] = xr.DataArray(np.arange(5))
        scene["raspberry"] = xr.DataArray(np.arange(5))
        #  deepcode ignore replace~keys~list~compare: This is on purpose
        assert make_cid(name="blueberry") in scene.keys()
        assert make_cid(name="blueberry") in scene
        assert "blueberry" in scene
        assert "blueberry" not in scene.keys()

    def test_delitem(self):
        """Test deleting an item."""
        scene = Scene()
        scene["1"] = xr.DataArray(np.arange(5),
                                  attrs={"wavelength": (0.1, 0.2, 0.3),
                                         "_satpy_id_keys": default_id_keys_config})
        scene["2"] = xr.DataArray(np.arange(5),
                                  attrs={"wavelength": (0.4, 0.5, 0.6),
                                         "_satpy_id_keys": default_id_keys_config})
        scene["3"] = xr.DataArray(np.arange(5),
                                  attrs={"wavelength": (0.7, 0.8, 0.9),
                                         "_satpy_id_keys": default_id_keys_config})
        del scene["1"]
        del scene["3"]
        del scene[0.45]
        assert not scene._wishlist
        assert not list(scene._datasets.keys())
        pytest.raises(KeyError, scene.__delitem__, 0.2)


def _create_coarest_finest_data_array(shape, area_def, attrs=None):
    data_arr = xr.DataArray(
        da.arange(math.prod(shape)).reshape(shape),
        attrs={
            "area": area_def,
        })
    if attrs:
        data_arr.attrs.update(attrs)
    return data_arr


def _create_coarsest_finest_area_def(shape, extents):
    from pyresample import AreaDefinition
    proj_str = "+proj=lcc +datum=WGS84 +ellps=WGS84 +lon_0=-95. +lat_0=25 +lat_1=25 +units=m +no_defs"
    area_def = AreaDefinition(
        "test",
        "test",
        "test",
        proj_str,
        shape[1],
        shape[0],
        extents,
    )
    return area_def


def _create_coarsest_finest_swath_def(shape, extents, name_suffix):
    from pyresample import SwathDefinition
    if len(shape) == 1:
        lons_arr = da.linspace(extents[0], extents[2], shape[0], dtype=np.float32)
        lats_arr = da.linspace(extents[1], extents[3], shape[0], dtype=np.float32)
    else:
        lons_arr = da.repeat(da.linspace(extents[0], extents[2], shape[1], dtype=np.float32)[None, :], shape[0], axis=0)
        lats_arr = da.repeat(da.linspace(extents[1], extents[3], shape[0], dtype=np.float32)[:, None], shape[1], axis=1)
    lons_data_arr = xr.DataArray(lons_arr, attrs={"name": f"longitude{name_suffix}"})
    lats_data_arr = xr.DataArray(lats_arr, attrs={"name": f"latitude1{name_suffix}"})
    return SwathDefinition(lons_data_arr, lats_data_arr)


class TestFinestCoarsestArea:
    """Test the Scene logic for finding the finest and coarsest area."""

    @pytest.mark.parametrize(
        ("coarse_area", "fine_area"),
        [
            (_create_coarsest_finest_area_def((2, 5), (1000.0, 1500.0, -1000.0, -1500.0)),
             _create_coarsest_finest_area_def((4, 10), (1000.0, 1500.0, -1000.0, -1500.0))),
            (_create_coarsest_finest_area_def((2, 5), (-1000.0, -1500.0, 1000.0, 1500.0)),
             _create_coarsest_finest_area_def((4, 10), (-1000.0, -1500.0, 1000.0, 1500.0))),
            (_create_coarsest_finest_swath_def((2, 5), (1000.0, 1500.0, -1000.0, -1500.0), "1"),
             _create_coarsest_finest_swath_def((4, 10), (1000.0, 1500.0, -1000.0, -1500.0), "1")),
            (_create_coarsest_finest_swath_def((5,), (1000.0, 1500.0, -1000.0, -1500.0), "1"),
             _create_coarsest_finest_swath_def((10,), (1000.0, 1500.0, -1000.0, -1500.0), "1")),
        ]
    )
    def test_coarsest_finest_area_different_shape(self, coarse_area, fine_area):
        """Test 'coarsest_area' and 'finest_area' methods for upright areas."""
        ds1 = _create_coarest_finest_data_array(coarse_area.shape, coarse_area, {"wavelength": (0.1, 0.2, 0.3)})
        ds2 = _create_coarest_finest_data_array(fine_area.shape, fine_area, {"wavelength": (0.4, 0.5, 0.6)})
        ds3 = _create_coarest_finest_data_array(fine_area.shape, fine_area, {"wavelength": (0.7, 0.8, 0.9)})
        scn = Scene()
        scn["1"] = ds1
        scn["2"] = ds2
        scn["3"] = ds3

        assert scn.coarsest_area() is coarse_area
        assert scn.finest_area() is fine_area
        assert scn.coarsest_area(["2", "3"]) is fine_area

    @pytest.mark.parametrize(
        ("area_def", "shifted_area"),
        [
            (_create_coarsest_finest_area_def((2, 5), (-1000.0, -1500.0, 1000.0, 1500.0)),
             _create_coarsest_finest_area_def((2, 5), (-900.0, -1400.0, 1100.0, 1600.0))),
            (_create_coarsest_finest_swath_def((2, 5), (-1000.0, -1500.0, 1000.0, 1500.0), "1"),
             _create_coarsest_finest_swath_def((2, 5), (-900.0, -1400.0, 1100.0, 1600.0), "2")),
        ],
    )
    def test_coarsest_finest_area_same_shape(self, area_def, shifted_area):
        """Test that two areas with the same shape are consistently returned.

        If two geometries (ex. two AreaDefinitions or two SwathDefinitions)
        have the same resolution (shape) but different
        coordinates, which one has the finer resolution would ultimately be
        determined by the semi-random ordering of the internal container of
        the Scene (a dict) if only pixel resolution was compared. This test
        makes sure that it is always the same object returned.

        """
        ds1 = _create_coarest_finest_data_array(area_def.shape, area_def)
        ds2 = _create_coarest_finest_data_array(area_def.shape, shifted_area)
        scn = Scene()
        scn["ds1"] = ds1
        scn["ds2"] = ds2
        course_area1 = scn.coarsest_area()

        scn = Scene()
        scn["ds2"] = ds2
        scn["ds1"] = ds1
        coarse_area2 = scn.coarsest_area()
        # doesn't matter what order they were added, this should be the same area
        assert coarse_area2 is course_area1


@pytest.mark.usefixtures("include_test_etc")
class TestComputePersist:
    """Test methods that compute the internal data in some way."""

    def test_compute_pass_through(self):
        """Test pass through of xarray compute."""
        import numpy as np
        scene = Scene(filenames=["fake1_1.txt"], reader="fake1")
        scene.load(["ds1"])
        scene = scene.compute()
        assert isinstance(scene["ds1"].data, np.ndarray)

    def test_persist_pass_through(self):
        """Test pass through of xarray persist."""
        from dask.array.utils import assert_eq
        scene = Scene(filenames=["fake1_1.txt"], reader="fake1")
        scene.load(["ds1"])
        scenep = scene.persist()
        assert_eq(scene["ds1"].data, scenep["ds1"].data)
        assert set(scenep["ds1"].data.dask).issubset(scene["ds1"].data.dask)
        assert len(scenep["ds1"].data.dask) == scenep["ds1"].data.npartitions

    def test_chunk_pass_through(self):
        """Test pass through of xarray chunk."""
        scene = Scene(filenames=["fake1_1.txt"], reader="fake1")
        scene.load(["ds1"])
        scene = scene.chunk(chunks=2)
        assert scene["ds1"].data.chunksize == (2, 2)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017-2023 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Tests CF-compliant DataArray creation."""
import numpy as np
import pytest
import xarray as xr

from satpy.tests.utils import make_dsq


def test_preprocess_dataarray_name():
    """Test saving an array to netcdf/cf where dataset name starting with a digit with prefix include orig name."""
    from satpy import Scene
    from satpy.cf.data_array import _preprocess_data_array_name

    scn = Scene()
    scn["1"] = xr.DataArray([1, 2, 3])
    dataarray = scn["1"]
    # If numeric_name_prefix is a string, test add the original_name attributes
    out_da = _preprocess_data_array_name(dataarray, numeric_name_prefix="TEST", include_orig_name=True)
    assert out_da.attrs["original_name"] == "1"

    # If numeric_name_prefix is empty string, False or None, test do not add original_name attributes
    with pytest.warns(UserWarning, match="Invalid NetCDF dataset name"):
        out_da = _preprocess_data_array_name(dataarray, numeric_name_prefix="", include_orig_name=True)
    assert "original_name" not in out_da.attrs

    with pytest.warns(UserWarning, match="Invalid NetCDF dataset name"):
        out_da = _preprocess_data_array_name(dataarray, numeric_name_prefix=False, include_orig_name=True)
    assert "original_name" not in out_da.attrs

    with pytest.warns(UserWarning, match="Invalid NetCDF dataset name"):
        out_da = _preprocess_data_array_name(dataarray, numeric_name_prefix=None, include_orig_name=True)
    assert "original_name" not in out_da.attrs


def test_make_cf_dataarray_lonlat():
    """Test correct CF encoding for area with lon/lat units."""
    from pyresample import create_area_def

    from satpy.cf.data_array import make_cf_data_array
    from satpy.coords import add_crs_xy_coords

    area = create_area_def("mavas", 4326, shape=(5, 5),
                           center=(0, 0), resolution=(1, 1))
    da = xr.DataArray(
        np.arange(25).reshape(5, 5),
        dims=("y", "x"),
        attrs={"area": area})
    da = add_crs_xy_coords(da, area)
    new_da = make_cf_data_array(da)
    assert new_da["x"].attrs["units"] == "degrees_east"
    assert new_da["y"].attrs["units"] == "degrees_north"


class TestCfDataArray:
    """Test creation of CF DataArray."""

    def test_make_cf_dataarray(self):
        """Test the conversion of a DataArray to a CF-compatible DataArray."""
        from satpy.cf.data_array import make_cf_data_array
        from satpy.tests.cf_tests._test_data import get_test_attrs
        from satpy.tests.utils import assert_dict_array_equality

        # Create set of test attributes
        attrs, attrs_expected, attrs_expected_flat = get_test_attrs()
        attrs["area"] = "some_area"
        attrs["prerequisites"] = [make_dsq(name="hej")]
        attrs["_satpy_id_name"] = "myname"

        # Adjust expected attributes
        expected_prereq = ("DataQuery(name='hej')")
        update = {"prerequisites": [expected_prereq], "long_name": attrs["name"]}

        attrs_expected.update(update)
        attrs_expected_flat.update(update)

        attrs_expected.pop("name")
        attrs_expected_flat.pop("name")

        # Create test data array
        arr = xr.DataArray(np.array([[1, 2], [3, 4]]), attrs=attrs, dims=("y", "x"),
                           coords={"y": [0, 1], "x": [1, 2], "acq_time": ("y", [3, 4])})

        # Test conversion to something cf-compliant
        res = make_cf_data_array(arr)
        np.testing.assert_array_equal(res["x"], arr["x"])
        np.testing.assert_array_equal(res["y"], arr["y"])
        np.testing.assert_array_equal(res["acq_time"], arr["acq_time"])
        assert res["x"].attrs == {"units": "m", "standard_name": "projection_x_coordinate"}
        assert res["y"].attrs == {"units": "m", "standard_name": "projection_y_coordinate"}
        assert_dict_array_equality(res.attrs, attrs_expected)

        # Test attribute kwargs
        res_flat = make_cf_data_array(arr, flatten_attrs=True, exclude_attrs=["int"])
        attrs_expected_flat.pop("int")
        assert_dict_array_equality(res_flat.attrs, attrs_expected_flat)

    def test_make_cf_dataarray_one_dimensional_array(self):
        """Test the conversion of an 1d DataArray to a CF-compatible DataArray."""
        from satpy.cf.data_array import make_cf_data_array

        arr = xr.DataArray(np.array([1, 2, 3, 4]), attrs={}, dims=("y",),
                           coords={"y": [0, 1, 2, 3], "acq_time": ("y", [0, 1, 2, 3])})
        _ = make_cf_data_array(arr)



"""Tests for the testing helper module."""

import numpy as np
import xarray as xr

from satpy import Scene
from satpy.area import get_area_def
from satpy.testing import fake_satpy_reading


def test_fake_reading(tmp_path):
    """Test that the fake reading context manager populates a scene."""
    input_files = [tmp_path / "my_input_file"]
    area = get_area_def("euro4")
    random = np.random.default_rng()
    somedata = xr.DataArray(random.uniform(size=area.shape), dims=["y", "x"])
    somedata.attrs["area"] = area

    channel = "VIS006"

    scene_dict = {channel: somedata}

    with fake_satpy_reading(scene_dict):
        scene = Scene(input_files, reader="dummy_reader")
        scene.load([channel])
    assert scene[channel] is somedata



#!/usr/bin/env python
# Copyright (c) 2016-2025 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Advance Baseline Imager reader base class for the Level 1b and l2+ reader."""

import datetime as dt
import logging
import math
from contextlib import suppress

import dask
import numpy as np
import xarray as xr
from pyresample import geometry

from satpy._compat import cached_property
from satpy.readers.core.file_handlers import BaseFileHandler
from satpy.readers.core.remote import open_file_or_filename
from satpy.utils import get_dask_chunk_size_in_bytes

logger = logging.getLogger(__name__)

PLATFORM_NAMES = {
    "g16": "GOES-16",
    "g17": "GOES-17",
    "g18": "GOES-18",
    "g19": "GOES-19",
    "goes16": "GOES-16",
    "goes17": "GOES-17",
    "goes18": "GOES-18",
}


class NC_ABI_BASE(BaseFileHandler):
    """Base reader for ABI L1B  L2+ NetCDF4 files."""

    def __init__(self, filename, filename_info, filetype_info, **kwargs):
        """Open the NetCDF file with xarray and prepare the Dataset for reading."""
        super(NC_ABI_BASE, self).__init__(filename, filename_info, filetype_info)

        platform_shortname = filename_info["platform_shortname"]
        self.platform_name = PLATFORM_NAMES.get(platform_shortname.lower())

        self.nlines = self.nc["y"].size
        self.ncols = self.nc["x"].size

        self.coords = {}

    @cached_property
    def nc(self):
        """Get the xarray dataset for this file."""
        chunk_bytes = self._chunk_bytes_for_resolution()
        with dask.config.set({"array.chunk-size": chunk_bytes}):
            f_obj = open_file_or_filename(self.filename)
            nc = xr.open_dataset(f_obj,
                                 decode_cf=True,
                                 mask_and_scale=False,
                                 chunks="auto")
        nc = self._rename_dims(nc)
        return nc

    def _chunk_bytes_for_resolution(self) -> int:
        """Get a best-guess optimal chunk size for resolution-based chunking.

        First a chunk size is chosen for the provided Dask setting `array.chunk-size`
        and then aligned with a hardcoded on-disk chunk size of 226. This is then
        adjusted to match the current resolution.

        This should result in 500 meter data having 4 times as many pixels per
        dask array chunk (2 in each dimension) as 1km data and 8 times as many
        as 2km data. As data is combined or upsampled geographically the arrays
        should not need to be rechunked. Care is taken to make sure that array
        chunks are aligned with on-disk file chunks at all resolutions, but at
        the cost of flexibility due to a hardcoded on-disk chunk size of 226
        elements per dimension.

        """
        num_high_res_elems_per_dim = math.sqrt(get_dask_chunk_size_in_bytes() / 4)  # 32-bit floats
        # assume on-disk chunk size of 226
        # this is true for all CSPP Geo GRB output (226 for all sectors) and full disk from other sources
        # 250 has been seen for AWS/CLASS CONUS, Mesoscale 1, and Mesoscale 2 files
        # we align this with 4 on-disk chunks at 500m, so it will be 2 on-disk chunks for 1km, and 1 for 2km
        high_res_elems_disk_aligned = round(max(num_high_res_elems_per_dim / (4 * 226), 1)) * (4 * 226)
        low_res_factor = int(self.filetype_info.get("resolution", 2000) // 500)
        res_elems_per_dim = int(high_res_elems_disk_aligned / low_res_factor)
        return (res_elems_per_dim ** 2) * 2  # 16-bit integers on disk

    @staticmethod
    def _rename_dims(nc):
        if "t" in nc.dims or "t" in nc.coords:
            nc = nc.rename({"t": "time"})
        if "goes_lat_lon_projection" in nc:
            with suppress(ValueError):
                nc = nc.rename({"lon": "x", "lat": "y"})
        return nc

    @property
    def sensor(self):
        """Get sensor name for current file handler."""
        return "abi"

    def __getitem__(self, item):
        """Wrap `self.nc[item]` for better floating point precision.

        Some datasets use a 32-bit float scaling factor like the 'x' and 'y'
        variables which causes inaccurate unscaled data values. This method
        forces the scale factor to a 64-bit float first.
        """
        data = self.nc[item]
        attrs = data.attrs

        data = self._adjust_data(data, item)

        data.attrs = attrs

        data = self._adjust_coords(data, item)

        return data

    def __contains__(self, item):
        """Check if the specified variable exists in this file."""
        return item in self.nc.variables

    def _adjust_data(self, data, item):
        """Adjust data with typing, scaling and filling."""
        factor = data.attrs.get("scale_factor", 1)
        offset = data.attrs.get("add_offset", 0)
        fill = data.attrs.get("_FillValue")
        unsigned = data.attrs.get("_Unsigned", None)

        def is_int(val):
            return np.issubdtype(val.dtype, np.integer) if hasattr(val, "dtype") else isinstance(val, int)

        # Ref. GOESR PUG-L1B-vol3, section 5.0.2 Unsigned Integer Processing
        if unsigned is not None and unsigned.lower() == "true":
            # cast the data from int to uint
            data = data.astype("u%s" % data.dtype.itemsize)

            if fill is not None:
                fill = fill.astype("u%s" % fill.dtype.itemsize)
        if fill is not None:
            # Some backends (h5netcdf) may return attributes as shape (1,)
            # arrays rather than shape () scalars, which according to the netcdf
            # documentation at <URL:https://www.unidata.ucar.edu
            # /software/netcdf/docs/netcdf_data_set_components.html#attributes>
            # is correct.
            if np.ndim(fill) > 0:
                fill = fill.item()
            if is_int(data) and is_int(factor) and is_int(offset):
                new_fill = fill
            else:
                new_fill = np.float32(np.nan)
            data = data.where(data != fill, new_fill)
        if factor != 1 and item in ("x", "y"):
            # be more precise with x/y coordinates
            # see get_area_def for more information
            data = data * np.round(float(factor), 6) + np.round(float(offset), 6)
        elif factor != 1:
            data = data * np.float32(factor) + np.float32(offset)
        return data

    def _adjust_coords(self, data, item):
        """Handle coordinates (and recursive fun)."""
        new_coords = {}
        # 'time' dimension causes issues in other processing
        # 'x_image' and 'y_image' are confusing to some users and unnecessary
        # 'x' and 'y' will be overwritten by base class AreaDefinition
        for coord_name in ("x_image", "y_image", "time", "x", "y"):
            if coord_name in data.coords:
                data = data.drop_vars(coord_name)
        if item in data.coords:
            self.coords[item] = data
        for coord_name in data.coords.keys():
            if coord_name not in self.coords:
                self.coords[coord_name] = self[coord_name]
            new_coords[coord_name] = self.coords[coord_name]
        data.coords.update(new_coords)
        return data

    def get_dataset(self, key, info):
        """Load a dataset."""
        raise NotImplementedError("Reader {} has not implemented get_dataset".format(self.name))

    def get_area_def(self, key):
        """Get the area definition of the data at hand."""
        if "goes_imager_projection" in self.nc:
            return self._get_areadef_fixedgrid(key)
        if "goes_lat_lon_projection" in self.nc:
            return self._get_areadef_latlon(key)
        raise ValueError("Unsupported projection found in the dataset")

    def _get_areadef_latlon(self, key):
        """Get the area definition of the data at hand."""
        projection = self.nc["goes_lat_lon_projection"]

        a = projection.attrs["semi_major_axis"]
        b = projection.attrs["semi_minor_axis"]
        fi = projection.attrs["inverse_flattening"]
        pm = projection.attrs["longitude_of_prime_meridian"]

        proj_ext = self.nc["geospatial_lat_lon_extent"]

        w_lon = proj_ext.attrs["geospatial_westbound_longitude"]
        e_lon = proj_ext.attrs["geospatial_eastbound_longitude"]
        n_lat = proj_ext.attrs["geospatial_northbound_latitude"]
        s_lat = proj_ext.attrs["geospatial_southbound_latitude"]

        lat_0 = proj_ext.attrs["geospatial_lat_center"]
        lon_0 = proj_ext.attrs["geospatial_lon_center"]

        area_extent = (w_lon, s_lat, e_lon, n_lat)
        proj_dict = {"proj": "latlong",
                     "lon_0": float(lon_0),
                     "lat_0": float(lat_0),
                     "a": float(a),
                     "b": float(b),
                     "fi": float(fi),
                     "pm": float(pm)}

        ll_area_def = geometry.AreaDefinition(
            self.nc.attrs.get("orbital_slot", "abi_geos"),
            self.nc.attrs.get("spatial_resolution", "ABI file area"),
            "abi_latlon",
            proj_dict,
            self.ncols,
            self.nlines,
            np.asarray(area_extent))

        return ll_area_def

    def _get_areadef_fixedgrid(self, key):
        """Get the area definition of the data at hand.

        Note this method takes special care to round and cast numbers to new
        data types so that the area definitions for different resolutions
        (different bands) should be equal. Without the special rounding in
        `__getitem__` and this method the area extents can be 0 to 1.0 meters
        off depending on how the calculations are done.

        """
        projection = self.nc["goes_imager_projection"]
        a = projection.attrs["semi_major_axis"]
        b = projection.attrs["semi_minor_axis"]
        h = projection.attrs["perspective_point_height"]

        lon_0 = projection.attrs["longitude_of_projection_origin"]
        sweep_axis = projection.attrs["sweep_angle_axis"][0]

        # compute x and y extents in m
        h = np.float64(h)
        x = self["x"]
        y = self["y"]
        x_l = x[0].values
        x_r = x[-1].values
        y_l = y[-1].values
        y_u = y[0].values
        x_half = (x_r - x_l) / (self.ncols - 1) / 2.
        y_half = (y_u - y_l) / (self.nlines - 1) / 2.
        area_extent = (x_l - x_half, y_l - y_half, x_r + x_half, y_u + y_half)
        area_extent = tuple(np.round(h * val, 6) for val in area_extent)

        proj_dict = {"proj": "geos",
                     "lon_0": float(lon_0),
                     "a": float(a),
                     "b": float(b),
                     "h": h,
                     "units": "m",
                     "sweep": sweep_axis}

        fg_area_def = geometry.AreaDefinition(
            self.nc.attrs.get("orbital_slot", "abi_geos"),
            self.nc.attrs.get("spatial_resolution", "ABI file area"),
            "abi_fixed_grid",
            proj_dict,
            self.ncols,
            self.nlines,
            np.asarray(area_extent))

        return fg_area_def

    @property
    def start_time(self):
        """Start time of the current file's observations."""
        return dt.datetime.strptime(self.nc.attrs["time_coverage_start"], "%Y-%m-%dT%H:%M:%S.%fZ")

    @property
    def end_time(self):
        """End time of the current file's observations."""
        return dt.datetime.strptime(self.nc.attrs["time_coverage_end"], "%Y-%m-%dT%H:%M:%S.%fZ")

    def spatial_resolution_to_number(self):
        """Convert the 'spatial_resolution' global attribute to meters."""
        res = self.nc.attrs["spatial_resolution"].split(" ")[0]
        if res.endswith("km"):
            res = int(float(res[:-2]) * 1000)
        elif res.endswith("m"):
            res = int(res[:-1])
        else:
            raise ValueError("Unexpected 'spatial_resolution' attribute '{}'".format(res))
        return res



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2014-2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Compact viirs format.

This is a reader for the Compact VIIRS format shipped on Eumetcast for the
VIIRS SDR. The format is compressed in multiple ways, notably by shipping only
tie-points for geographical data. The interpolation of this data is done using
dask operations, so it should be relatively performant.

For more information on this format, the reader can refer to the
`Compact VIIRS SDR Product Format User Guide` that can be found on this EARS_ page.

.. _EARS: https://www.eumetsat.int/media/45988

"""

import datetime as dt
import logging
from contextlib import suppress

import dask.array as da
import h5py
import numpy as np
import xarray as xr

from satpy.readers.core.file_handlers import BaseFileHandler
from satpy.readers.core.utils import np2str
from satpy.utils import angle2xyz, get_legacy_chunk_size, lonlat2xyz, xyz2angle, xyz2lonlat

CHUNK_SIZE = get_legacy_chunk_size()
_channels_dict = {"M01": "M1",
                  "M02": "M2",
                  "M03": "M3",
                  "M04": "M4",
                  "M05": "M5",
                  "M06": "M6",
                  "M07": "M7",
                  "M08": "M8",
                  "M09": "M9",
                  "M10": "M10",
                  "M11": "M11",
                  "M12": "M12",
                  "M13": "M13",
                  "M14": "M14",
                  "M15": "M15",
                  "M16": "M16",
                  "DNB": "DNB"}

logger = logging.getLogger(__name__)

c = 299792458  # m.s-1
h = 6.6260755e-34  # m2kg.s-1
k = 1.380658e-23  # m2kg.s-2.K-1

short_names = {"NPP": "Suomi-NPP",
               "J01": "NOAA-20",
               "J02": "NOAA-21"}


class VIIRSCompactFileHandler(BaseFileHandler):
    """A file handler class for VIIRS compact format."""

    def __init__(self, filename, filename_info, filetype_info):
        """Initialize the reader."""
        super(VIIRSCompactFileHandler, self).__init__(filename, filename_info,
                                                      filetype_info)
        self.h5f = h5py.File(self.filename, "r")
        self.finfo = filename_info
        self.lons = None
        self.lats = None
        if filetype_info["file_type"] == "compact_m":
            self.ch_type = "MOD"
        elif filetype_info["file_type"] == "compact_dnb":
            self.ch_type = "DNB"
        else:
            raise IOError("Compact Viirs file type not recognized.")

        geo_data = self.h5f["Data_Products"]["VIIRS-%s-GEO" % self.ch_type]["VIIRS-%s-GEO_Gran_0" % self.ch_type]
        self.min_lat = geo_data.attrs["South_Bounding_Coordinate"].item()
        self.max_lat = geo_data.attrs["North_Bounding_Coordinate"].item()
        self.min_lon = geo_data.attrs["West_Bounding_Coordinate"].item()
        self.max_lon = geo_data.attrs["East_Bounding_Coordinate"].item()

        self.switch_to_cart = ((abs(self.max_lon - self.min_lon) > 90)
                               or (max(abs(self.min_lat), abs(self.max_lat)) > 60))

        self.scans = self.h5f["All_Data"]["NumberOfScans"][0]
        self.geography = self.h5f["All_Data"]["VIIRS-%s-GEO_All" % self.ch_type]

        for key in self.h5f["All_Data"].keys():
            if key.startswith("VIIRS") and key.endswith("SDR_All"):
                channel = key.split("-")[1]
                break

        # This supposes there is only one tiepoint zone in the track direction.
        channel_path = f"All_Data/VIIRS-{channel}-SDR_All"
        self.scan_size = self.h5f[channel_path].attrs["TiePointZoneSizeTrack"].item()
        self.track_offset = self.h5f[channel_path].attrs["PixelOffsetTrack"][()]
        self.scan_offset = self.h5f[channel_path].attrs["PixelOffsetScan"][()]

        try:
            self.group_locations = self.geography["TiePointZoneGroupLocationScanCompact"][()]
        except KeyError:
            self.group_locations = [0]

        self.tpz_sizes = da.from_array(self.h5f[channel_path].attrs["TiePointZoneSizeScan"], chunks=1)
        if len(self.tpz_sizes.shape) == 2:
            if self.tpz_sizes.shape[1] != 1:
                raise NotImplementedError("Can't handle 2 dimensional tiepoint zones.")
            self.tpz_sizes = self.tpz_sizes.squeeze(1)
        self.nb_tiepoint_zones = self.geography["NumberOfTiePointZonesScan"][()]
        self.c_align = da.from_array(self.geography["AlignmentCoefficient"],
                                     chunks=tuple(self.nb_tiepoint_zones))
        self.c_exp = da.from_array(self.geography["ExpansionCoefficient"],
                                   chunks=tuple(self.nb_tiepoint_zones))
        self.nb_tiepoint_zones = da.from_array(self.nb_tiepoint_zones, chunks=1)
        self._expansion_coefs = None

        self.cache = {}

        self.mda = {}
        short_name = np2str(self.h5f.attrs["Platform_Short_Name"])
        self.mda["platform_name"] = short_names.get(short_name, short_name)
        self.mda["sensor"] = "viirs"

    def __del__(self):
        """Close file handlers when we are done."""
        with suppress(OSError):
            self.h5f.close()

    def get_dataset(self, key, info):
        """Load a dataset."""
        logger.debug("Reading %s.", key["name"])
        if key["name"] in _channels_dict:
            m_data = self.read_dataset(key, info)
        else:
            m_data = self.read_geo(key, info)
        m_data.attrs.update(info)
        m_data.attrs["rows_per_scan"] = self.scan_size
        return m_data

    def get_bounding_box(self):
        """Get the bounding box of the data."""
        for key in self.h5f["Data_Products"].keys():
            if key.startswith("VIIRS") and key.endswith("GEO"):
                lats = self.h5f["Data_Products"][key][key + "_Gran_0"].attrs["G-Ring_Latitude"][()]
                lons = self.h5f["Data_Products"][key][key + "_Gran_0"].attrs["G-Ring_Longitude"][()]
                break
        else:
            raise KeyError("Cannot find bounding coordinates!")
        return lons.ravel(), lats.ravel()

    @property
    def start_time(self):
        """Get the start time."""
        return self.finfo["start_time"]

    @property
    def end_time(self):
        """Get the end time."""
        end_time = dt.datetime.combine(self.start_time.date(),
                                    self.finfo["end_time"].time())
        if end_time < self.start_time:
            end_time += dt.timedelta(days=1)
        return end_time

    def read_geo(self, key, info):
        """Read angles."""
        pairs = {("satellite_azimuth_angle", "satellite_zenith_angle"):
                 ("SatelliteAzimuthAngle", "SatelliteZenithAngle"),
                 ("solar_azimuth_angle", "solar_zenith_angle"):
                 ("SolarAzimuthAngle", "SolarZenithAngle"),
                 ("dnb_solar_azimuth_angle", "dnb_solar_zenith_angle"):
                 ("SolarAzimuthAngle", "SolarZenithAngle"),
                 ("dnb_lunar_azimuth_angle", "dnb_lunar_zenith_angle"):
                 ("LunarAzimuthAngle", "LunarZenithAngle"),
                 }
        if self.lons is None or self.lats is None:
            self.lons, self.lats = self.navigate()
        for pair, fkeys in pairs.items():
            if key["name"] in pair:
                if (self.cache.get(pair[0]) is None
                        or self.cache.get(pair[1]) is None):
                    angles = self.angles(*fkeys)
                    self.cache[pair[0]], self.cache[pair[1]] = angles
                if key["name"] == pair[0]:
                    return xr.DataArray(self.cache[pair[0]], name=key["name"],
                                        attrs=self.mda, dims=("y", "x"))
                else:
                    return xr.DataArray(self.cache[pair[1]], name=key["name"],
                                        attrs=self.mda, dims=("y", "x"))

        if info.get("standard_name") in ["latitude", "longitude"]:
            mda = self.mda.copy()
            mda.update(info)
            if info["standard_name"] == "longitude":
                return xr.DataArray(self.lons, attrs=mda, dims=("y", "x"))
            else:
                return xr.DataArray(self.lats, attrs=mda, dims=("y", "x"))

        if key["name"] == "dnb_moon_illumination_fraction":
            mda = self.mda.copy()
            mda.update(info)
            return xr.DataArray(da.from_array(self.geography["MoonIllumFraction"]),
                                attrs=info)

    def read_dataset(self, dataset_key, info):
        """Read a dataset."""
        h5f = self.h5f
        channel = _channels_dict[dataset_key["name"]]
        chan_dict = dict([(key.split("-")[1], key)
                          for key in h5f["All_Data"].keys()
                          if key.startswith("VIIRS")])

        h5rads = h5f["All_Data"][chan_dict[channel]]["Radiance"]
        chunks = h5rads.chunks or CHUNK_SIZE
        rads = xr.DataArray(da.from_array(h5rads, chunks=chunks),
                            name=dataset_key["name"],
                            dims=["y", "x"]).astype(np.float32)
        h5attrs = h5rads.attrs
        scans = h5f["All_Data"]["NumberOfScans"][0]
        rads = rads[:scans * 16, :]
        rads = rads.where(rads <= 65526)
        try:
            rads = xr.where(rads <= h5attrs["Threshold"],
                            rads * h5attrs["RadianceScaleLow"] +
                            h5attrs["RadianceOffsetLow"],
                            rads * h5attrs["RadianceScaleHigh"] +
                            h5attrs["RadianceOffsetHigh"])
        except (KeyError, AttributeError):
            logger.info("Missing attribute for scaling of %s.", channel)
            pass
        unit = "W m-2 sr-1 μm-1"
        if dataset_key["calibration"] == "counts":
            raise NotImplementedError("Can't get counts from this data")
        if dataset_key["calibration"] in ["reflectance", "brightness_temperature"]:
            # do calibrate
            try:
                # First guess: VIS or NIR data
                a_vis = h5attrs["EquivalentWidth"]
                b_vis = h5attrs["IntegratedSolarIrradiance"]
                dse = h5attrs["EarthSunDistanceNormalised"]
                rads *= 100 * np.pi * a_vis / b_vis * (dse**2)
                unit = "%"
            except KeyError:
                # Maybe it's IR data?
                try:
                    a_ir = h5attrs["BandCorrectionCoefficientA"]
                    b_ir = h5attrs["BandCorrectionCoefficientB"]
                    lambda_c = h5attrs["CentralWaveLength"]
                    rads *= 1e6
                    rads = (h * c) / (k * lambda_c *
                                      np.log(1 +
                                             (2 * h * c ** 2) /
                                             ((lambda_c ** 5) * rads)))
                    rads *= a_ir
                    rads += b_ir
                    unit = "K"
                except KeyError:
                    logger.warning("Calibration failed.")

        elif dataset_key["calibration"] != "radiance":
            raise ValueError("Calibration parameter should be radiance, "
                             "reflectance or brightness_temperature")
        rads = rads.clip(min=0)
        rads.attrs = self.mda
        rads.attrs["units"] = unit
        return rads

    def expand_angle_and_nav(self, arrays):
        """Expand angle and navigation datasets."""
        res = []
        for array in arrays:
            res.append(da.map_blocks(expand, array[:, :, np.newaxis], self.expansion_coefs,
                                     scans=self.scans, scan_size=self.scan_size,
                                     dtype=array.dtype, drop_axis=2, chunks=self.expansion_coefs.chunks[:-1]))
        return res

    @property
    def expansion_coefs(self):
        """Compute the expansion coefficients."""
        if self._expansion_coefs is not None:
            return self._expansion_coefs
        v_track = (np.arange(self.scans * self.scan_size) % self.scan_size + self.track_offset) / self.scan_size
        self.tpz_sizes = self.tpz_sizes.persist()
        self.nb_tiepoint_zones = self.nb_tiepoint_zones.persist()
        col_chunks = (self.tpz_sizes * self.nb_tiepoint_zones).compute()
        self._expansion_coefs = da.map_blocks(get_coefs, self.c_align, self.c_exp, self.tpz_sizes,
                                              self.nb_tiepoint_zones,
                                              v_track=v_track, scans=self.scans, scan_size=self.scan_size,
                                              scan_offset=self.scan_offset,
                                              dtype=np.float64, new_axis=[0, 2],
                                              chunks=(self.scans * self.scan_size, tuple(col_chunks), 4))

        return self._expansion_coefs

    def navigate(self):
        """Generate the navigation datasets."""
        chunks = self._get_geographical_chunks()
        lon = da.from_array(self.geography["Longitude"], chunks=chunks)
        lat = da.from_array(self.geography["Latitude"], chunks=chunks)
        if self.switch_to_cart:
            arrays = lonlat2xyz(lon, lat)
        else:
            arrays = (lon, lat)

        expanded = self.expand_angle_and_nav(arrays)
        if self.switch_to_cart:
            return xyz2lonlat(*expanded)

        return expanded

    def _get_geographical_chunks(self):
        shape = self.geography["Longitude"].shape
        horizontal_chunks = (self.nb_tiepoint_zones + 1).compute()
        chunks = (shape[0], tuple(horizontal_chunks))
        return chunks

    def angles(self, azi_name, zen_name):
        """Generate the angle datasets."""
        chunks = self._get_geographical_chunks()

        azi = self.geography[azi_name]
        zen = self.geography[zen_name]

        switch_to_cart = ((np.max(azi) - np.min(azi) > 5)
                          or (np.min(zen) < 10)
                          or (max(abs(self.min_lat), abs(self.max_lat)) > 80))

        azi = da.from_array(azi, chunks=chunks)
        zen = da.from_array(zen, chunks=chunks)

        if switch_to_cart:
            arrays = convert_from_angles(azi, zen)
        else:
            arrays = (azi, zen)

        expanded = self.expand_angle_and_nav(arrays)
        if switch_to_cart:
            return convert_to_angles(*expanded)

        return expanded


def convert_from_angles(azi, zen):
    """Convert the angles to cartesian coordinates."""
    x, y, z, = angle2xyz(azi, zen)
    # Conversion to ECEF is recommended by the provider, but no significant
    # difference has been seen.
    # x, y, z = (-np.sin(lon) * x + np.cos(lon) * y,
    #            -np.sin(lat) * np.cos(lon) * x - np.sin(lat) * np.sin(lon) * y + np.cos(lat) * z,
    #            np.cos(lat) * np.cos(lon) * x + np.cos(lat) * np.sin(lon) * y + np.sin(lat) * z)
    return x, y, z


def convert_to_angles(x, y, z):
    """Convert the cartesian coordinates to angles."""
    # Conversion to ECEF is recommended by the provider, but no significant
    # difference has been seen.
    # x, y, z = (-np.sin(lon) * x - np.sin(lat) * np.cos(lon) * y + np.cos(lat) * np.cos(lon) * z,
    #            np.cos(lon) * x - np.sin(lat) * np.sin(lon) * y + np.cos(lat) * np.sin(lon) * z,
    #            np.cos(lat) * y + np.sin(lat) * z)
    azi, zen = xyz2angle(x, y, z, acos=True)
    return azi, zen


def get_coefs(c_align, c_exp, tpz_size, nb_tpz, v_track, scans, scan_size, scan_offset):
    """Compute the coeffs in numpy domain."""
    nties = nb_tpz.item()
    tpz_size = tpz_size.item()
    v_scan = (np.arange(nties * tpz_size) % tpz_size + scan_offset) / tpz_size
    s_scan, s_track = np.meshgrid(v_scan, v_track)
    s_track = s_track.reshape(scans, scan_size, nties, tpz_size)
    s_scan = s_scan.reshape(scans, scan_size, nties, tpz_size)

    c_align = c_align[np.newaxis, np.newaxis, :, np.newaxis]
    c_exp = c_exp[np.newaxis, np.newaxis, :, np.newaxis]

    a_scan = s_scan + s_scan * (1 - s_scan) * c_exp + s_track * (
        1 - s_track) * c_align
    a_track = s_track
    coef_a = (1 - a_track) * (1 - a_scan)
    coef_b = (1 - a_track) * a_scan
    coef_d = a_track * (1 - a_scan)
    coef_c = a_track * a_scan
    res = np.stack([coef_a, coef_b, coef_c, coef_d], axis=4).reshape(scans * scan_size, -1, 4)
    return res


def expand(data, coefs, scans, scan_size):
    """Perform the expansion in numpy domain."""
    data = data.reshape(data.shape[:-1])

    coefs = coefs.reshape(scans, scan_size, data.shape[1] - 1, -1, 4)

    coef_a = coefs[:, :, :, :, 0]
    coef_b = coefs[:, :, :, :, 1]
    coef_c = coefs[:, :, :, :, 2]
    coef_d = coefs[:, :, :, :, 3]

    corner_coefficients = (coef_a, coef_b, coef_c, coef_d)
    fdata = _interpolate_data(data, corner_coefficients, scans)
    return fdata.reshape(scans * scan_size, -1)


def _interpolate_data(data, corner_coefficients, scans):
    """Interpolate the data using the provided coefficients."""
    coef_a, coef_b, coef_c, coef_d = corner_coefficients
    data_a = data[:scans * 2:2, np.newaxis, :-1, np.newaxis]
    data_b = data[:scans * 2:2, np.newaxis, 1:, np.newaxis]
    data_c = data[1:scans * 2:2, np.newaxis, 1:, np.newaxis]
    data_d = data[1:scans * 2:2, np.newaxis, :-1, np.newaxis]
    fdata = (coef_a * data_a + coef_b * data_b + coef_d * data_d + coef_c * data_c)
    return fdata


def expand_arrays(arrays,
                  scans,
                  c_align,
                  c_exp,
                  scan_size=16,
                  tpz_size=16,
                  nties=200,
                  track_offset=0.5,
                  scan_offset=0.5):
    """Expand *data* according to alignment and expansion."""
    nties = nties.item()
    tpz_size = tpz_size.item()
    s_scan, s_track = da.meshgrid(da.arange(nties * tpz_size),
                                  da.arange(scans * scan_size))
    s_track = (s_track.reshape(scans, scan_size, nties, tpz_size) % scan_size
               + track_offset) / scan_size
    s_scan = (s_scan.reshape(scans, scan_size, nties, tpz_size) % tpz_size
              + scan_offset) / tpz_size

    a_scan = s_scan + s_scan * (1 - s_scan) * c_exp + s_track * (
        1 - s_track) * c_align
    a_track = s_track
    expanded = []
    coef_a = (1 - a_track) * (1 - a_scan)
    coef_b = (1 - a_track) * a_scan
    coef_d = a_track * (1 - a_scan)
    coef_c = a_track * a_scan
    corner_coefficients = (coef_a, coef_b, coef_c, coef_d)
    for data in arrays:
        fdata = _interpolate_data(data, corner_coefficients, scans)
        expanded.append(fdata.reshape(scans * scan_size, nties * tpz_size))
    return expanded
